"""
See COPYRIGHT.md for copyright information.

EBA (2.3), EIOPA (2.0.0) Filing Rules Validation
"""
import os, sys
import regex as re
from arelle import PluginManager
from arelle import ModelDocument, XbrlConst, XmlUtil, UrlUtil, LeiUtil
from arelle.HashUtil import md5hash, Md5Sum
from arelle.ModelDtsObject import ModelConcept, ModelType, ModelLocator, ModelResource
from arelle.ModelFormulaObject import Aspect
from arelle.ModelObject import ModelObject
from arelle.ModelRelationshipSet import ModelRelationshipSet
from arelle.ModelValue import qname, qnameEltPfxName
from arelle.ValidateUtr import ValidateUtr
from arelle.Version import authorLabel, copyrightLabel
from arelle.XbrlConst import qnEnumerationItemTypes
from arelle.ModelInstanceObject import ModelFact
from arelle.utils.Contexts import getDuplicateContextGroups
import regex as re
from lxml import etree
from collections import defaultdict

qnFIndicators = qname("{http://www.eurofiling.info/xbrl/ext/filing-indicators}find:fIndicators")
qnFilingIndicator = qname("{http://www.eurofiling.info/xbrl/ext/filing-indicators}find:filingIndicator")
qnPercentItemType = qname("{http://www.xbrl.org/dtr/type/numeric}num:percentItemType")
qnPureItemType = qname("{http://www.xbrl.org/2003/instance}xbrli:pureItemType")
qnMetReportingCurrency = qname("{http://eiopa.europa.eu/xbrl/s2md/dict/met}met:ei1930")
integerItemTypes = {"integerItemType", "nonPositiveIntegerItemType", "negativeIntegerItemType",
                    "longItemType", "intItemType", "shortItemType", "byteItemType",
                    "nonNegativeIntegerItemType", "unsignedLongItemType", "unsignedIntItemType",
                    "unsignedShortItemType", "unsignedByteItemType", "positiveIntegerItemType"}
schemaRefDatePattern = re.compile(r".*/([0-9]{4}-[01][0-9]-[0-3][0-9])/.*")

s_2_18_c_a_met = {
    """
        in templates S.06.02, SE.06.02, S.08.01, S.08.02,S.11.01 and E.01.01,
        data points with the data type 'monetary' shall be expressed in units
        with at least two decimals:
            select distinct mem.MemberXBRLCode from mOrdinateCategorisation oc
            inner join mAxisOrdinate ao on ao.OrdinateID = oc.OrdinateID
            inner join mTableAxis ta on ta.AxisID = ao.AxisID
            inner join mTable t on t.TableID = ta.TableID
            inner join mMember mem on mem.MemberID = oc.MemberID
            inner join mMetric met on met.CorrespondingMemberID = mem.MemberID and met.DataType = 'Monetary'
            where (t.TableCode like 'S.06.02%' or t.TableCode like 'SE.06.02%' or t.TableCode like 'S.08.01%' or t.TableCode like 'S.08.02%' or t.TableCode like 'S.11.01%' or t.TableCode like 'E.01.01%') and mem.MemberXBRLCode not like 's2hd_met%'
            order by t.TableCode;
    """
    "s2md_met:mi1088", "s2md_met:mi1096", "s2md_met:mi1101", "s2md_met:mi1110",
    "s2md_met:mi1112", "s2md_met:mi1115", "s2md_met:mi1117", "s2md_met:mi1126",
    "s2md_met:mi1127", "s2md_met:mi1128", "s2md_met:mi1131"}

CANONICAL_PREFIXES = {
    "http://www.xbrl.org/2003/iso4217": "iso4217",
    "http://www.xbrl.org/2003/linkbase": "link",
    "http://xbrl.org/2006/xbrldi": "xbrldi",
    "http://www.xbrl.org/2003/instance": "xbrli",
    "http://www.w3.org/1999/xlink": "xlink"}

def dislosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("EBA", "EBA"),
            ("EIOPA", "EIOPA"))

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(__file__), "config.xml")

def validateSetup(val, parameters=None, *args, **kwargs):
    val.validateEBA = val.validateDisclosureSystem and getattr(val.disclosureSystem, "EBA", False)
    val.validateEIOPA = val.validateDisclosureSystem and getattr(val.disclosureSystem, "EIOPA", False)
    if not (val.validateEBA or val.validateEIOPA):
        return

    val.validateUTR = False # do not use default UTR validation, it's at error level and not streamable
    val.utrValidator = ValidateUtr(val.modelXbrl,
                                   "WARNING",  # EBA specifies SHOULD on UTR validation
                                   "EBA.2.23") # override utre error-severity message code

    val.isEIOPAfullVersion = val.isEIOPA_2_0_1 = False
    modelDocument = val.modelXbrl.modelDocument
    if modelDocument.type == ModelDocument.Type.INSTANCE:
        for doc, docRef in modelDocument.referencesDocument.items():
            if "href" in docRef.referenceTypes:
                if docRef.referringModelObject.localName == "schemaRef":
                    _match = schemaRefDatePattern.match(doc.uri)
                    if _match:
                        val.isEIOPAfullVersion = _match.group(1) > "2015-02-28"
                        val.isEIOPA_2_0_1 = _match.group(1) >= "2015-10-21"
                        break
                    else:
                        val.modelXbrl.error( ("EBA.S.1.5.a/EBA.S.1.5.b", "EIOPA.S.1.5.a/EIOPA.S.1.5.b"),
                                        _('The link:schemaRef element in submitted instances MUST resolve to the full published entry point URL, this schemaRef is missing date portion: %(schemaRef)s.'),
                                        modelObject=modelDocument, schemaRef=doc.uri)

    val.qnDimAF = val.qnDimOC = val.qnCAx1 = None
    _nsmap = val.modelXbrl.modelDocument.xmlRootElement.nsmap

    if val.isEIOPA_2_0_1:
        _hasPiInstanceGenerator = False
        for pi in modelDocument.processingInstructions:
            if pi.target == "instance-generator":
                _hasPiInstanceGenerator = True
                if not all(pi.get(attr) for attr in ("id", "version", "creationdate")):
                    val.modelXbrl.warning("EIOPA.S.2.23",
                                          _('The processing instruction instance-generator SHOULD contain attributes "id", "version" and "creationdate".'),
                                          modelObject=modelDocument)
        if not _hasPiInstanceGenerator:
            val.modelXbrl.warning("EIOPA.S.2.23",
                                  _('The instance SHOULD include a processing instruction "instance-generator".'),
                                  modelObject=modelDocument)

        val.qnDimAF = qname("s2c_dim:AF", _nsmap)
        val.qnDimOC = qname("s2c_dim:OC", _nsmap)
        val.qnCAx1 = qname("s2c_CA:x1", _nsmap)
    elif val.validateEBA:
        val.eba_qnDimCUS = qname("eba_dim:CUS", _nsmap)
        val.eba_qnDimCCA = qname("eba_dim:CCA", _nsmap)
        val.eba_qnCAx1 = qname("eba_CA:x1", _nsmap)


    val.prefixNamespace = {}
    val.namespacePrefix = {}
    val.idObjects = {}

    val.typedDomainQnames = set()
    val.typedDomainElements = set()
    for modelConcept in val.modelXbrl.qnameConcepts.values():
        if modelConcept.isTypedDimension:
            typedDomainElement = modelConcept.typedDomainElement
            if isinstance(typedDomainElement, ModelConcept):
                val.typedDomainQnames.add(typedDomainElement.qname)
                val.typedDomainElements.add(typedDomainElement)

    val.filingIndicators = {}
    val.numFilingIndicatorTuples = 0

    val.cntxEntities = set()
    val.cntxDates = defaultdict(set)
    val.unusedCntxIDs = set()
    val.unusedUnitIDs = set()
    val.currenciesUsed = {}
    val.reportingCurrency = None
    val.namespacePrefixesUsed = defaultdict(set)
    val.prefixesUnused = set()
    for prefix, ns in _nsmap.items():
        val.prefixesUnused.add(prefix)
        val.namespacePrefixesUsed[ns].add(prefix)
    val.firstFactObjectIndex = sys.maxsize
    val.firstFact = None
    val.footnotesRelationshipSet = ModelRelationshipSet(val.modelXbrl, "XBRL-footnotes")
    # re-init batch flag to enable more than one context/unit validation sessions for the same instance.
    # (note that this monkey-patching would give trouble on two concurrent validation sessions of the same instance)
    for cntx in val.modelXbrl.contexts.values():
        if hasattr(cntx, "_batchChecked"):
            cntx._batchChecked = False
    for unit in val.modelXbrl.units.values():
        if hasattr(unit, "_batchChecked"):
            unit._batchChecked = False

def prefixUsed(val, ns, prefix):
    val.namespacePrefixesUsed[ns].add(prefix)
    for _prefix in val.namespacePrefixesUsed[ns]:
        val.prefixesUnused.discard(_prefix)

def validateStreamingFacts(val, factsToCheck, *args, **kwargs):
    if not (val.validateEBA or val.validateEIOPA):
        return True
    validateFacts(val, factsToCheck)

def validateFacts(val, factsToCheck):
    # may be called in streaming batches or all at end (final) if not streaming

    modelXbrl = val.modelXbrl
    modelDocument = modelXbrl.modelDocument

    # note EBA 2.1 is in ModelDocument.py

    timelessDatePattern = re.compile(r"\s*([0-9]{4})-([0-9]{2})-([0-9]{2})\s*$")
    for cntx in modelXbrl.contexts.values():
        if getattr(cntx, "_batchChecked", False):
            continue # prior streaming batch already checked
        cntx._batchChecked = True
        val.cntxEntities.add(cntx.entityIdentifier)
        dateElts = XmlUtil.descendants(cntx, XbrlConst.xbrli, ("startDate","endDate","instant"))
        if any(not timelessDatePattern.match(e.textValue) for e in dateElts):
            modelXbrl.error(("EBA.2.10","EIOPA.2.10"),
                    _('Period dates must be whole dates without time or timezone: %(dates)s.'),
                    modelObject=cntx, dates=", ".join(e.text for e in dateElts))
        if cntx.isForeverPeriod:
            modelXbrl.error(("EBA.2.11","EIOPA.N.2.11"),
                    _('Forever context period is not allowed.'),
                    modelObject=cntx)
        elif cntx.isStartEndPeriod:
            modelXbrl.error(("EBA.2.13","EIOPA.N.2.11"),
                    _('Start-End (flow) context period is not allowed.'),
                    modelObject=cntx)
        elif cntx.isInstantPeriod:
            # cannot pass context object to final() below, for error logging, if streaming mode
            val.cntxDates[cntx.instantDatetime].add(modelXbrl if getattr(val.modelXbrl, "isStreamingMode", False)
                                                    else cntx)
        if cntx.hasSegment:
            modelXbrl.error(("EBA.2.14","EIOPA.N.2.14"),
                _("Contexts MUST NOT contain xbrli:segment values: %(cntx)s.'"),
                modelObject=cntx, cntx=cntx.id)
        if cntx.nonDimValues("scenario"):
            modelXbrl.error(("EBA.2.15","EIOPA.S.2.15" if val.isEIOPAfullVersion else "EIOPA.N.2.15"),
                _("Contexts MUST NOT contain non-dimensional xbrli:scenario values: %(cntx)s.'"),
                modelObject=cntx, cntx=cntx.id,
                messageCodes=("EBA.2.15","EIOPA.N.2.15","EIOPA.S.2.15"))
        val.unusedCntxIDs.add(cntx.id)
        if val.isEIOPA_2_0_1 and len(cntx.id) > 128:
            modelXbrl.warning("EIOPA.S.2.6",
                _("Contexts IDs SHOULD be short: %(cntx)s.'"),
                modelObject=cntx, cntx=cntx.id)

    for unit in modelXbrl.units.values():
        if getattr(unit, "_batchChecked", False):
            continue # prior streaming batch already checked
        unit._batchChecked = True
        val.unusedUnitIDs.add(unit.id)

    factsByQname = defaultdict(set) # top level for this
    for f in factsToCheck:
        factsByQname[f.qname].add(f)
        val.unusedCntxIDs.discard(f.contextID)
        val.unusedUnitIDs.discard(f.unitID)
        if f.objectIndex < val.firstFactObjectIndex:
            val.firstFactObjectIndex = f.objectIndex
            val.firstFact = f


    for fIndicators in factsByQname[qnFIndicators]:
        val.numFilingIndicatorTuples += 1
        for fIndicator in fIndicators.modelTupleFacts:
            _value = (getattr(fIndicator, "xValue", None) or fIndicator.value) # use validated xValue if DTS else value for skipDTS
            _filed = fIndicator.get("{http://www.eurofiling.info/xbrl/ext/filing-indicators}filed", "true") in ("true", "1")
            if _value in val.filingIndicators:
                modelXbrl.error(("EBA.1.6.1", "EIOPA.1.6.1"),
                        _('Multiple filing indicators facts for indicator %(filingIndicator)s.'),
                        modelObject=(fIndicator, val.filingIndicators[_value]), filingIndicator=_value)
                if _filed and not val.filingIndicators[_value]:
                    val.filingIndicators[_value] = _filed #set to filed if any of the multiple indicators are filed=true
            else: # not a duplicate filing indicator
                val.filingIndicators[_value] = _filed
            val.unusedCntxIDs.discard(fIndicator.contextID)
            cntx = fIndicator.context
            if cntx is not None and (cntx.hasSegment or cntx.hasScenario):
                modelXbrl.error("EIOPA.N.1.6.d" if val.isEIOPAfullVersion else "EIOPA.S.1.6.d",
                        _('Filing indicators must not contain segment or scenario elements %(filingIndicator)s.'),
                        modelObject=fIndicator, filingIndicator=_value)
        # Using model object id's is not accurate in case of edition
        prevObj = fIndicators.getprevious()
        while prevObj is not None:
            if isinstance(prevObj, ModelFact) and prevObj.qname != qnFIndicators:
                modelXbrl.warning("EIOPA.1.6.2",
                              _('Filing indicators should precede first fact %(firstFact)s.'),
                              modelObject=(fIndicators, val.firstFact), firstFact=val.firstFact.qname)
                break
            prevObj = prevObj.getprevious()

    if val.isEIOPAfullVersion:
        for fIndicator in factsByQname[qnFilingIndicator]:
            if fIndicator.getparent().qname == XbrlConst.qnXbrliXbrl:
                _isPos = fIndicator.get("{http://www.eurofiling.info/xbrl/ext/filing-indicators}filed", "true") in ("true", "1")
                _value = (getattr(fIndicator, "xValue", None) or fIndicator.value) # use validated xValue if DTS else value for skipDTS
                modelXbrl.error("EIOPA.1.6.a" if _isPos else "EIOPA.1.6.b",
                        _('Filing indicators must be in a tuple %(filingIndicator)s.'),
                        modelObject=fIndicator, filingIndicator=_value,
                        messageCodes=("EIOPA.1.6.a", "EIOPA.1.6.b"))

    otherFacts = {} # (contextHash, unitHash, xmlLangHash) : fact
    nilFacts = []
    stringFactsWithXmlLang = []
    nonMonetaryNonPureFacts = []
    for qname, facts in factsByQname.items():
        for f in facts:
            if f.qname == qnFIndicators or f.qname == qnFIndicators:
                continue # skip root-level and non-root-level filing indicators
            if modelXbrl.skipDTS:
                c = f.qname.localName[0]
                isNumeric = c in ('m', 'p', 'r', 'i')
                isMonetary = c == 'm'
                isInteger = c == 'i'
                isPercent = c == 'p'
                isString = c == 's'
                isEnum = c == 'e'
            else:
                concept = f.concept
                if concept is not None:
                    isNumeric = concept.isNumeric
                    isMonetary = concept.isMonetary
                    isInteger = concept.baseXbrliType in integerItemTypes
                    isPercent = concept.typeQname in (qnPercentItemType, qnPureItemType)
                    isString = concept.baseXbrliType in ("stringItemType", "normalizedStringItemType")
                    isEnum = concept.typeQname in qnEnumerationItemTypes
                else:
                    isNumeric = isString = isEnum = False # error situation
            k = (f.getparent().objectIndex,
                 f.qname,
                 f.context.contextDimAwareHash if f.context is not None else None,
                 f.unit.hash if f.unit is not None else None,
                 hash(f.xmlLang))
            if k not in otherFacts:
                otherFacts[k] = {f}
            else:
                matches = [o
                           for o in otherFacts[k]
                           if (f.getparent().objectIndex == o.getparent().objectIndex and
                               f.qname == o.qname and
                               f.context.isEqualTo(o.context) if f.context is not None and o.context is not None else True) and
                               # (f.unit.isEqualTo(o.unit) if f.unit is not None and o.unit is not None else True) and
                              (f.xmlLang == o.xmlLang)]
                if matches:
                    contexts = [f.contextID] + [o.contextID for o in matches]
                    modelXbrl.error(("EBA.2.16", "EIOPA.S.2.16" if val.isEIOPAfullVersion else "EIOPA.S.2.16.a"),
                                    _('Facts are duplicates %(fact)s contexts %(contexts)s.'),
                                    modelObject=[f] + matches, fact=f.qname, contexts=', '.join(contexts),
                                    messageCodes=("EBA.2.16", "EIOPA.S.2.16", "EIOPA.S.2.16.a"))
                else:
                    otherFacts[k].add(f)
            if isNumeric:
                if f.precision:
                    modelXbrl.error(("EBA.2.17", "EIOPA.2.18.a"),
                        _("Numeric fact %(fact)s of context %(contextID)s has a precision attribute '%(precision)s'"),
                        modelObject=f, fact=f.qname, contextID=f.contextID, precision=f.precision)
                if f.decimals and not f.isNil: # in XbrlDpmSqlDB for 2_0_1
                    if f.decimals == "INF":
                        if not val.isEIOPAfullVersion:
                            modelXbrl.error("EIOPA.S.2.18.f",
                                _("Monetary fact %(fact)s of context %(contextID)s has a decimal attribute INF: '%(decimals)s'"),
                                modelObject=f, fact=f.qname, contextID=f.contextID, decimals=f.decimals)
                    else:
                        try:
                            xValue = f.xValue
                            dec = int(f.decimals)
                            if isMonetary:
                                if val.isEIOPA_2_0_1:
                                    _absXvalue = abs(xValue)
                                    if str(f.qname) in s_2_18_c_a_met:
                                        dMin = 2
                                    elif _absXvalue >= 100000000:
                                        dMin = -4
                                    elif 100000000 > _absXvalue >= 1000000:
                                        dMin = -3
                                    elif 1000000 > _absXvalue >= 1000:
                                        dMin = -2
                                    else:
                                        dMin = -1
                                    if dMin > dec:
                                        modelXbrl.error("EIOPA.S.2.18.c",
                                            _("Monetary fact %(fact)s of context %(contextID)s has a decimals attribute less than minimum %(minimumDecimals)s: '%(decimals)s'"),
                                            modelObject=f, fact=f.qname, contextID=f.contextID, minimumDecimals=dMin, decimals=f.decimals)
                                elif dec < -3:
                                    modelXbrl.error(("EBA.2.18","EIOPA.S.2.18.c"),
                                        _("Monetary fact %(fact)s of context %(contextID)s has a decimals attribute < -3: '%(decimals)s'"),
                                        modelObject=f, fact=f.qname, contextID=f.contextID, decimals=f.decimals)
                                else: # apply dynamic decimals check
                                    if  -.1 < xValue < .1: dMin = 2
                                    elif -1 < xValue < 1: dMin = 1
                                    elif -10 < xValue < 10: dMin = 0
                                    elif -100 < xValue < 100: dMin = -1
                                    elif -1000 < xValue < 1000: dMin = -2
                                    else: dMin = -3
                                    if dMin > dec:
                                        modelXbrl.warning("EIOPA:factDecimalsWarning",
                                            _("Monetary fact %(fact)s of context %(contextID)s value %(value)s has an imprecise decimals attribute: %(decimals)s, minimum is %(mindec)s"),
                                            modelObject=f, fact=f.qname, contextID=f.contextID, value=xValue, decimals=f.decimals, mindec=dMin)
                            elif isInteger:
                                if dec != 0:
                                    modelXbrl.error(("EBA.2.18","EIOPA.S.2.18.d"),
                                        _("Integer fact %(fact)s of context %(contextID)s has a decimals attribute \u2260 0: '%(decimals)s'"),
                                        modelObject=f, fact=f.qname, contextID=f.contextID, decimals=f.decimals)
                            elif isPercent:
                                if dec < 4:
                                    modelXbrl.error(("EBA.2.18","EIOPA.S.2.18.e"),
                                        _("Percent fact %(fact)s of context %(contextID)s has a decimals attribute < 4: '%(decimals)s'"),
                                        modelObject=f, fact=f.qname, contextID=f.contextID, decimals=f.decimals)
                                if val.isEIOPA_2_0_1 and xValue > 1:
                                    modelXbrl.warning(("EIOPA.3.2.b"),
                                        _("Percent fact %(fact)s of context %(contextID)s appears to be over 100% = 1.0: '%(value)s'"),
                                        modelObject=f, fact=f.qname, contextID=f.contextID, value=xValue)
                            else:
                                if -.001 < xValue < .001: dMin = 4
                                elif -.01 < xValue < .01: dMin = 3
                                elif -.1 < xValue < .1: dMin = 2
                                elif  -1 < xValue < 1: dMin = 1
                                else: dMin = 0
                                if dMin > dec:
                                    modelXbrl.warning("EIOPA:factDecimalsWarning",
                                        _("Numeric fact %(fact)s of context %(contextID)s value %(value)s has an imprecise decimals attribute: %(decimals)s, minimum is %(mindec)s"),
                                        modelObject=f, fact=f.qname, contextID=f.contextID, value=xValue, decimals=f.decimals, mindec=dMin)
                        except (AttributeError, ValueError):
                            pass # should have been reported as a schema error by loader
                        '''' (not intended by EBA 2.18, paste here is from EFM)
                        if not f.isNil and getattr(f,"xValid", 0) == 4:
                            try:
                                insignificance = insignificantDigits(f.xValue, decimals=f.decimals)
                                if insignificance: # if not None, returns (truncatedDigits, insiginficantDigits)
                                    modelXbrl.error(("EFM.6.05.37", "GFM.1.02.26"),
                                        _("Fact %(fact)s of context %(contextID)s decimals %(decimals)s value %(value)s has nonzero digits in insignificant portion %(insignificantDigits)s."),
                                        modelObject=f1, fact=f1.qname, contextID=f1.contextID, decimals=f1.decimals,
                                        value=f1.xValue, truncatedDigits=insignificance[0], insignificantDigits=insignificance[1])
                            except (ValueError,TypeError):
                                modelXbrl.error(("EBA.2.18"),
                                    _("Fact %(fact)s of context %(contextID)s decimals %(decimals)s value %(value)s causes Value Error exception."),
                                    modelObject=f1, fact=f1.qname, contextID=f1.contextID, decimals=f1.decimals, value=f1.value)
                        '''
                unit = f.unit
                if unit is not None:
                    if isMonetary:
                        if unit.measures[0]:
                            _currencyMeasure = unit.measures[0][0]
                            if val.isEIOPA_2_0_1 and f.context is not None:
                                if f.context.dimMemberQname(val.qnDimAF) == val.qnCAx1 and val.qnDimOC in f.context.qnameDims:
                                    _ocCurrency = f.context.dimMemberQname(val.qnDimOC).localName
                                    if _currencyMeasure.localName != _ocCurrency:
                                        modelXbrl.error("EIOPA.3.1",
                                            _("There MUST be only one currency but metric %(metric)s reported OC dimension currency %(ocCurrency)s differs from unit currency: %(unitCurrency)s."),
                                            modelObject=f, metric=f.qname, ocCurrency=_ocCurrency, unitCurrency=_currencyMeasure.localName)
                                else:
                                    val.currenciesUsed[_currencyMeasure] = unit
                            elif val.validateEBA and f.context is not None:
                                if f.context.dimMemberQname(val.eba_qnDimCCA) == val.eba_qnCAx1 and val.eba_qnDimCUS in f.context.qnameDims:
                                    currency = f.context.dimMemberQname(val.eba_qnDimCUS).localName
                                    if _currencyMeasure.localName != currency:
                                        modelXbrl.error("EBA.3.1",
                                            _("There MUST be only one currency but metric %(metric)s reported CCA dimension currency %(currency)s differs from unit currency: %(unitCurrency)s."),
                                            modelObject=f, metric=f.qname, currency=currency, unitCurrency=_currencyMeasure.localName)
                                else:
                                    val.currenciesUsed[_currencyMeasure] = unit
                            else:
                                val.currenciesUsed[_currencyMeasure] = unit
                    elif not unit.isSingleMeasure or unit.measures[0][0] != XbrlConst.qnXbrliPure:
                        nonMonetaryNonPureFacts.append(f)
            if isEnum:
                _eQn = getattr(f,"xValue", None) or qnameEltPfxName(f, f.value)
                if _eQn:
                    prefixUsed(val, _eQn.namespaceURI, _eQn.prefix)
                    if val.isEIOPA_2_0_1 and f.qname.localName == "ei1930":
                        val.reportingCurrency = _eQn.localName
            elif isString:
                if f.xmlLang: # requires disclosureSystem to NOT specify default language
                    stringFactsWithXmlLang.append(f)

            if f.isNil:
                nilFacts.append(f)

            if val.footnotesRelationshipSet.fromModelObject(f):
                modelXbrl.warning("EIOPA.S.19",
                    _("Fact %(fact)s of context %(contextID)s has footnotes.'"),
                    modelObject=f, fact=f.qname, contextID=f.contextID)

    if nilFacts:
        modelXbrl.error(("EBA.2.19", "EIOPA.S.2.19"),
                _('Nil facts MUST NOT be present in the instance: %(nilFacts)s.'),
                modelObject=nilFacts, nilFacts=", ".join(str(f.qname) for f in nilFacts))
    if stringFactsWithXmlLang:
        modelXbrl.warning("EIOPA.2.20", # not reported for EBA
                          _("String facts reporting xml:lang (not saved by T4U, not round-tripped): '%(factsWithLang)s'"),
                          modelObject=stringFactsWithXmlLang, factsWithLang=", ".join(set(str(f.qname) for f in stringFactsWithXmlLang)))
    if nonMonetaryNonPureFacts:
        modelXbrl.error(("EBA.3.2","EIOPA.3.2.a"),
                        _("Non monetary (numeric) facts MUST use the pure unit: '%(langLessFacts)s'"),
                        modelObject=nonMonetaryNonPureFacts, langLessFacts=", ".join(set(str(f.qname) for f in nonMonetaryNonPureFacts)))

    val.utrValidator.validateFacts() # validate facts for UTR at logLevel WARNING

    unitHashes = {}
    for unit in modelXbrl.units.values():
        h = unit.hash
        if h in unitHashes and unit.isEqualTo(unitHashes[h]):
            modelXbrl.warning("EBA.2.21",
                _("Duplicate units SHOULD NOT be reported, units %(unit1)s and %(unit2)s have same measures.'"),
                modelObject=(unit, unitHashes[h]), unit1=unit.id, unit2=unitHashes[h].id)
            if not getattr(modelXbrl, "isStreamingMode", False):
                modelXbrl.error("EIOPA.2.21",
                    _("Duplicate units MUST NOT be reported, units %(unit1)s and %(unit2)s have same measures.'"),
                    modelObject=(unit, unitHashes[h]), unit1=unit.id, unit2=unitHashes[h].id)
        else:
            unitHashes[h] = unit
        for _measures in unit.measures:
            for _measure in _measures:
                prefixUsed(val, _measure.namespaceURI, _measure.prefix)

    del unitHashes

    if not getattr(modelXbrl, "isStreamingMode", False):
        for contexts in getDuplicateContextGroups(modelXbrl):
            modelXbrl.log("WARNING" if val.isEIOPAfullVersion else "ERROR",
                "EIOPA.S.2.7.b",
                _("Duplicate contexts MUST NOT be reported, contexts %(cntx1)s and %(cntx2)s are equivalent.'"),
                modelObject=contexts, cntx1=contexts[0].id, cntx2=contexts[1].id)
    for cntx in modelXbrl.contexts.values():
        for _dim in cntx.qnameDims.values():
            _dimQn = _dim.dimensionQname
            prefixUsed(val, _dimQn.namespaceURI, _dimQn.prefix)
            if _dim.isExplicit:
                _memQn = _dim.memberQname
            else:
                _memQn = _dim.typedMember.qname
            if _memQn:
                prefixUsed(val, _memQn.namespaceURI, _memQn.prefix)

    for elt in modelDocument.xmlRootElement.iter():
        if isinstance(elt, ModelObject): # skip comments and processing instructions
            prefixUsed(val, elt.qname.namespaceURI, elt.qname.prefix)
            for attrTag in elt.keys():
                if attrTag.startswith("{"):
                    _prefix, _NS, _localName = XmlUtil.clarkNotationToPrefixNsLocalname(elt, attrTag, isAttribute=True)
                    if _prefix:
                        prefixUsed(val, _NS, _prefix)
        elif val.isEIOPA_2_0_1:
            if elt.tag in ("{http://www.w3.org/2001/XMLSchema}documentation", "{http://www.w3.org/2001/XMLSchema}annotation"):
                modelXbrl.error("EIOPA.2.5",
                    _("xs:documentation element found, all relevant business data MUST only be contained in contexts, units, schemaRef and facts."),
                    modelObject=modelDocument)
            elif isinstance(elt, etree._Comment):
                modelXbrl.error("EIOPA.2.5",
                    _("XML comment found, all relevant business data MUST only be contained in contexts, units, schemaRef and facts: %(comment)s"),
                    modelObject=modelDocument, comment=elt.text)

def validateNonStreamingFinish(val, *args, **kwargs):
    # non-streaming EBA checks, ignore when streaming (first all from ValidateXbrl.py)
    if not getattr(val.modelXbrl, "isStreamingMode", False):
        final(val)

def validateStreamingFinish(val, *args, **kwargs):
    final(val)  # always finish validation when streaming

def final(val):
    if not (val.validateEBA or val.validateEIOPA):
        return

    modelXbrl = val.modelXbrl
    modelDocument = modelXbrl.modelDocument

    _statusMsg = _("validating {0} filing rules").format(val.disclosureSystem.name)
    modelXbrl.profileActivity()
    modelXbrl.modelManager.showStatus(_statusMsg)

    if modelDocument.type == ModelDocument.Type.INSTANCE and (val.validateEBA or val.validateEIOPA):

        if not modelDocument.uri.endswith(".xbrl"):
            modelXbrl.warning("EBA.1.1",
                    _('XBRL instance documents SHOULD use the extension ".xbrl" but it is "%(extension)s"'),
                    modelObject=modelDocument, extension=os.path.splitext(modelDocument.basename)[1])
            modelXbrl.error("EIOPA.S.1.1.a",
                    _('XBRL instance documents MUST use the extension ".xbrl" but it is "%(extension)s"'),
                    modelObject=modelDocument, extension=os.path.splitext(modelDocument.basename)[1])
        if val.isEIOPA_2_0_1: _encodings = ("UTF-8", "utf-8-sig")
        else: _encodings = ("utf-8", "UTF-8", "utf-8-sig")
        if modelDocument.documentEncoding not in _encodings:
            modelXbrl.error(("EBA.1.4", "EIOPA.1.4"),
                    _('XBRL instance documents MUST use "UTF-8" encoding but is "%(xmlEncoding)s"'),
                    modelObject=modelDocument, xmlEncoding=modelDocument.documentEncoding)

        schemaRefElts = []
        schemaRefFileNames = []
        for doc, docRef in modelDocument.referencesDocument.items():
            if "href" in docRef.referenceTypes:
                if docRef.referringModelObject.localName == "schemaRef":
                    schemaRefElts.append(docRef.referringModelObject)
                    schemaRefFileNames.append(doc.basename)
                    if not UrlUtil.isAbsolute(doc.uri):
                        modelXbrl.error(("EBA.2.2", "EIOPA.S.1.5.a" if val.isEIOPAfullVersion else "EIOPA.S.1.5.b"),
                                _('The link:schemaRef element in submitted instances MUST resolve to the full published entry point URL: %(url)s.'),
                                modelObject=docRef.referringModelObject, url=doc.uri,
                                messageCodes=("EBA.2.2", "EIOPA.S.1.5.a","EIOPA.S.1.5.b"))
                elif docRef.referringModelObject.localName == "linkbaseRef":
                    modelXbrl.error(("EBA.2.3","EIOPA.S.1.5.a"),
                            _('The link:linkbaseRef element is not allowed: %(fileName)s.'),
                            modelObject=docRef.referringModelObject, fileName=doc.basename)
        _numSchemaRefs = len(XmlUtil.children(modelDocument.xmlRootElement, XbrlConst.link, "schemaRef"))
        if _numSchemaRefs > 1:
            modelXbrl.error(("EIOPA.S.1.5.a", "EBA.1.5"),
                    _('XBRL instance documents MUST reference only one entry point schema but %(numEntryPoints)s were found: %(entryPointNames)s'),
                    modelObject=modelDocument, numEntryPoints=_numSchemaRefs, entryPointNames=', '.join(sorted(schemaRefFileNames)))
        ### check entry point names appropriate for filing indicator (DPM DB?)

        if len(schemaRefElts) != 1:
            modelXbrl.error("EBA.2.3",
                    _('Any reported XBRL instance document MUST contain only one xbrli:xbrl/link:schemaRef node, but %(entryPointCount)s.'),
                    modelObject=schemaRefElts, entryPointCount=len(schemaRefElts))
        # non-streaming EBA checks
        if not getattr(modelXbrl, "isStreamingMode", False):
            val.qnReportedCurrency = None
            if val.isEIOPA_2_0_1 and qnMetReportingCurrency in modelXbrl.factsByQname:
                for _multiCurrencyFact in modelXbrl.factsByQname[qnMetReportingCurrency]:
                    # multi-currency fact
                    val.qnReportedCurrency = _multiCurrencyFact.xValue
                    break

            validateFacts(val, modelXbrl.facts)

            # check sum of fact md5s (otherwise checked in streaming process)
            xbrlFactsCheckVersion = None
            expectedSumOfFactMd5s = None
            for pi in modelDocument.xmlRootElement.getchildren():
                if isinstance(pi, etree._ProcessingInstruction) and pi.target == "xbrl-facts-check":
                    _match = re.search("([\\w-]+)=[\"']([^\"']+)[\"']", pi.text)
                    if _match:
                        _matchGroups = _match.groups()
                        if len(_matchGroups) == 2:
                            if _matchGroups[0] == "version":
                                xbrlFactsCheckVersion = _matchGroups[1]
                            elif _matchGroups[0] == "sum-of-fact-md5s":
                                try:
                                    expectedSumOfFactMd5s = Md5Sum(_matchGroups[1])
                                except ValueError:
                                    modelXbrl.error("EIOPA:xbrlFactsCheckError",
                                            _("Invalid sum-of-md5s %(sumOfMd5)s"),
                                            modelObject=modelXbrl, sumOfMd5=_matchGroups[1])
            if xbrlFactsCheckVersion and expectedSumOfFactMd5s:
                sumOfFactMd5s = Md5Sum()
                for f in modelXbrl.factsInInstance:
                    sumOfFactMd5s += f.md5sum
                if sumOfFactMd5s != expectedSumOfFactMd5s:
                    modelXbrl.warning("EIOPA:xbrlFactsCheckWarning",
                            _("XBRL facts sum of md5s expected %(expectedMd5)s not matched to actual sum %(actualMd5Sum)s"),
                            modelObject=modelXbrl, expectedMd5=expectedSumOfFactMd5s, actualMd5Sum=sumOfFactMd5s)
                else:
                    modelXbrl.info("info",
                            _("Successful XBRL facts sum of md5s."),
                            modelObject=modelXbrl)

        if any(badError in modelXbrl.errors
               for badError in ("EBA.2.1", "EIOPA.2.1", "EIOPA.S.1.5.a/EIOPA.S.1.5.b")):
            pass # skip checking filingIndicators if bad errors
        elif not val.filingIndicators:
            modelXbrl.error(("EBA.1.6", "EIOPA.1.6.a"),
                    _('Missing filing indicators.  Reported XBRL instances MUST include appropriate (positive) filing indicator elements'),
                    modelObject=modelDocument)
        elif all(filed == False for filed in val.filingIndicators.values()):
            modelXbrl.error(("EBA.1.6", "EIOPA.1.6.a"),
                    _('All filing indicators are filed="false".  Reported XBRL instances MUST include appropriate (positive) filing indicator elements'),
                    modelObject=modelDocument)

        if val.numFilingIndicatorTuples > 1:
            modelXbrl.warning(("EBA.1.6.2", "EIOPA.1.6.2"),
                    _('Multiple filing indicators tuples when not in streaming mode (info).'),
                    modelObject=modelXbrl.factsByQname[qnFIndicators])

        if len(val.cntxDates) > 1:
            modelXbrl.error(("EBA.2.13","EIOPA.2.13"),
                    _('Contexts must have the same date: %(dates)s.'),
                    # when streaming values are no longer available, but without streaming they can be logged
                    modelObject=set(_cntx for _cntxs in val.cntxDates.values() for _cntx in _cntxs),
                    dates=', '.join(XmlUtil.dateunionValue(_dt, subtractOneDay=True)
                                                           for _dt in val.cntxDates.keys()))

        if val.unusedCntxIDs:
            if val.isEIOPA_2_0_1:
                modelXbrl.error("EIOPA.2.7",
                        _('Unused xbrli:context nodes MUST NOT be present in the instance: %(unusedContextIDs)s.'),
                        modelObject=[modelXbrl.contexts[unusedCntxID] for unusedCntxID in val.unusedCntxIDs if unusedCntxID in modelXbrl.contexts],
                        unusedContextIDs=", ".join(sorted(val.unusedCntxIDs)))
            else:
                modelXbrl.warning(("EBA.2.7", "EIOPA.2.7"),
                        _('Unused xbrli:context nodes SHOULD NOT be present in the instance: %(unusedContextIDs)s.'),
                        modelObject=[modelXbrl.contexts[unusedCntxID] for unusedCntxID in val.unusedCntxIDs if unusedCntxID in modelXbrl.contexts],
                        unusedContextIDs=", ".join(sorted(val.unusedCntxIDs)))

        if len(val.cntxEntities) > 1:
            modelXbrl.error(("EBA.2.9", "EIOPA.2.9"),
                    _('All entity identifiers and schemes MUST be the same, %(count)s found: %(entities)s.'),
                    modelObject=modelDocument, count=len(val.cntxEntities),
                    entities=", ".join(sorted(str(cntxEntity) for cntxEntity in val.cntxEntities)))

        for _scheme, _LEI in val.cntxEntities:
            if (_scheme in ("http://standards.iso.org/iso/17442", "http://standard.iso.org/iso/17442", "LEI") or
                (not val.isEIOPAfullVersion and _scheme == "PRE-LEI")):
                if _scheme == "http://standard.iso.org/iso/17442":
                    modelXbrl.warning(("EBA.3.6", "EIOPA.S.2.8.c"),
                        _("Warning, context has entity scheme %(scheme)s should be plural: http://standards.iso.org/iso/17442."),
                        modelObject=modelDocument, scheme=_scheme)
                result = LeiUtil.checkLei(_LEI)
                if result == LeiUtil.LEI_INVALID_LEXICAL:
                    modelXbrl.error("EIOPA.S.2.8.c",
                                    _("Context has lexically invalid LEI %(lei)s."),
                                    modelObject=modelDocument, lei=_LEI)
                elif result == LeiUtil.LEI_INVALID_CHECKSUM:
                    modelXbrl.error("EIOPA.S.2.8.c",
                                    _("Context has LEI checksum error in %(lei)s."),
                                    modelObject=modelDocument, lei=_LEI)
            elif _scheme == "SC":
                pass # anything is ok for Specific Code
            else:
                modelXbrl.error("EIOPA.S.2.8.c",
                    _("Context has unrecognized entity scheme %(scheme)s."),
                    modelObject=modelDocument, scheme=_scheme)

        if val.unusedUnitIDs:
            if val.isEIOPA_2_0_1:
                modelXbrl.error("EIOPA.2.22",
                        _('Unused xbrli:unit nodes MUST NOT be present in the instance: %(unusedUnitIDs)s.'),
                        modelObject=[modelXbrl.units[unusedUnitID] for unusedUnitID in val.unusedUnitIDs if unusedUnitID in modelXbrl.units],
                        unusedUnitIDs=", ".join(sorted(val.unusedUnitIDs)))
            else:
                modelXbrl.warning(("EBA.2.22", "EIOPA.2.22"),
                        _('Unused xbrli:unit nodes SHOULD NOT be present in the instance: %(unusedUnitIDs)s.'),
                        modelObject=[modelXbrl.units[unusedUnitID] for unusedUnitID in val.unusedUnitIDs if unusedUnitID in modelXbrl.units],
                        unusedUnitIDs=", ".join(sorted(val.unusedUnitIDs)))

        if len(val.currenciesUsed) > 1:
            modelXbrl.error(("EBA.3.1","EIOPA.3.1"),
                _("There MUST be only one currency but %(numCurrencies)s were found: %(currencies)s.'"),
                modelObject=val.currenciesUsed.values(), numCurrencies=len(val.currenciesUsed), currencies=", ".join(str(c) for c in val.currenciesUsed.keys()))

        elif val.isEIOPA_2_0_1 and any(_measure.localName != val.reportingCurrency for _measure in val.currenciesUsed.keys()):
            modelXbrl.error("EIOPA.3.1",
                _("There MUST be only one currency but reporting currency %(reportingCurrency)s differs from unit currencies: %(currencies)s.'"),
                modelObject=val.currenciesUsed.values(), reportingCurrency=val.reportingCurrency, currencies=", ".join(str(c) for c in val.currenciesUsed.keys()))

        if val.prefixesUnused:
            modelXbrl.warning(("EBA.3.4", "EIOPA.3.4"),
                _("There SHOULD be no unused prefixes but these were declared: %(unusedPrefixes)s.'"),
                modelObject=modelDocument, unusedPrefixes=', '.join(sorted(val.prefixesUnused)))
        for ns, prefixes in val.namespacePrefixesUsed.items():
            nsDocs = modelXbrl.namespaceDocs.get(ns)
            if nsDocs:
                for nsDoc in nsDocs:
                    nsDocPrefix = XmlUtil.xmlnsprefix(nsDoc.xmlRootElement, ns)
                    if any(prefix != nsDocPrefix for prefix in prefixes if prefix is not None):
                        modelXbrl.warning(("EBA.3.5", "EIOPA.3.5"),
                            _("Prefix for namespace %(namespace)s is %(declaredPrefix)s but these were found %(foundPrefixes)s"),
                            modelObject=modelDocument, namespace=ns, declaredPrefix=nsDocPrefix, foundPrefixes=', '.join(sorted(prefixes - {None})))
            elif ns in CANONICAL_PREFIXES and any(prefix != CANONICAL_PREFIXES[ns] for prefix in prefixes if prefix is not None):
                modelXbrl.warning(("EBA.3.5", "EIOPA.3.5"),
                    _("Prefix for namespace %(namespace)s is %(declaredPrefix)s but these were found %(foundPrefixes)s"),
                    modelObject=modelDocument, namespace=ns, declaredPrefix=CANONICAL_PREFIXES[ns], foundPrefixes=', '.join(sorted(prefixes - {None})))

    modelXbrl.profileActivity(_statusMsg, minTimeToShow=0.0)
    modelXbrl.modelManager.showStatus(None)

    del val.prefixNamespace, val.namespacePrefix, val.idObjects, val.typedDomainElements
    del val.utrValidator, val.firstFact, val.footnotesRelationshipSet

__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate EBA, EIOPA',
    'version': '1.2',
    'description': 'EBA (2.3), EIOPA (2.0.0) Filing Rules Validation.',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'DisclosureSystem.Types': dislosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
    'Validate.XBRL.Start': validateSetup,
    'Validate.XBRL.Finally': validateNonStreamingFinish,
    'Streaming.ValidateFacts': validateStreamingFacts,
    'Streaming.ValidateFinish': validateStreamingFinish,
}



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import *
try:
    from tkinter.ttk import *
except ImportError:
    from ttk import *
import os
from arelle import (ViewWinTree, ModelDocument, XmlUtil)
from arelle.ViewUtil import sortCountExpected

def viewTests(modelXbrl, tabWin):
    view = ViewTests(modelXbrl, tabWin)
    modelXbrl.modelManager.showStatus(_("viewing Tests"))
    view.treeView["columns"] = ("name", "readMeFirst", "infoset", "status", "call", "test", "expected", "actual")
    view.treeView.column("#0", width=150, anchor="w")
    view.treeView.heading("#0", text="ID")
    view.treeView.column("name", width=150, anchor="w")
    view.treeView.heading("name", text="Name")
    view.treeView.column("readMeFirst", width=75, anchor="w")
    view.treeView.heading("readMeFirst", text="ReadMeFirst")
    view.treeView.column("infoset", width=75, anchor="w")
    view.treeView.heading("infoset", text="Infoset File")
    view.treeView.column("status", width=80, anchor="w")
    view.treeView.heading("status", text="Status")
    view.treeView.column("call", width=150, anchor="w")
    view.treeView.heading("call", text="Call")
    view.treeView.column("test", width=100, anchor="w")
    view.treeView.heading("test", text="Test")
    view.treeView.column("expected", width=100, anchor="w")
    view.treeView.heading("expected", text="Expected")
    view.treeView.column("actual", width=100, anchor="w")
    view.treeView.heading("actual",  text="Actual")
    view.isTransformRegistry = False
    modelDocument = modelXbrl.modelDocument
    if modelXbrl.modelDocument.type in (ModelDocument.Type.REGISTRY, ModelDocument.Type.REGISTRYTESTCASE):
        if modelXbrl.modelDocument.xmlRootElement.namespaceURI == "http://xbrl.org/2011/conformance-rendering/transforms":
            view.treeView["displaycolumns"] = ("status", "call", "test", "expected", "actual")
            view.isTransformRegistry = True
        else:
            view.treeView["displaycolumns"] = ("name", "readMeFirst", "status", "call", "test", "expected", "actual")
    elif modelXbrl.modelDocument.type == ModelDocument.Type.XPATHTESTSUITE:
        view.treeView["displaycolumns"] = ("name", "readMeFirst", "status", "call", "test", "expected", "actual")
    else:
        # check if infoset needed
        if modelDocument.type in (ModelDocument.Type.TESTCASESINDEX, ModelDocument.Type.REGISTRY):
            hasInfoset = any(getattr(refDoc, "outpath", None)  for refDoc in modelDocument.referencesDocument)
        else:
            hasInfoset = bool(getattr(modelDocument, "outpath", None))
        view.treeView["displaycolumns"] = (("name", "readMeFirst") +
                                           ( ("infoset",) if hasInfoset else () ) +
                                           ( "status", "expected", "actual"))

    menu = view.contextMenu()
    view.menuAddExpandCollapse()
    view.menuAddClipboard()
    view.id = 0
    view.viewTestcaseIndexElement(modelDocument, "")
    view.blockSelectEvent = 1
    view.blockViewModelObject = 0
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

class ViewTests(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin):
        super(ViewTests, self).__init__(modelXbrl, tabWin, "Tests", True)

    def viewTestcaseIndexElement(self, modelDocument, parentNode, parentNodeText=None):
        self.id += 1
        if modelDocument.type in (ModelDocument.Type.TESTCASESINDEX, ModelDocument.Type.REGISTRY):
            nodeText = os.path.basename(modelDocument.uri)
            if nodeText == parentNodeText: # may be same name, index.xml, use directory name instead
                nodeText = os.path.basename(os.path.dirname(modelDocument.uri))
            node = self.treeView.insert(parentNode, "end", modelDocument.objectId(self.id),
                                        text=nodeText, tags=("odd",))
            self.id += 1;
            # sort test cases by uri
            testcases = []
            for referencedDocument, _ref in sorted(modelDocument.referencesDocument.items(),
                                                   key=lambda i:i[1].referringModelObject.objectIndex if i[1] else 0):
                if referencedDocument.type in (ModelDocument.Type.TESTCASESINDEX, ModelDocument.Type.REGISTRY):
                    self.viewTestcaseIndexElement(referencedDocument, node, nodeText)
                else:
                    testcases.append((referencedDocument.uri, referencedDocument.objectId()))
            testcases.sort()
            for i, testcaseTuple in enumerate(testcases):
                self.viewTestcase(self.modelXbrl.modelObject(testcaseTuple[1]), node, i)
        elif modelDocument.type in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
            self.viewTestcase(modelDocument, parentNode, 1)
        elif modelDocument.type == ModelDocument.Type.XPATHTESTSUITE:
            for i, elt in enumerate(modelDocument.xmlRootElement.iterchildren(tag="{http://www.w3.org/2005/02/query-test-XQTSCatalog}test-group")):
                self.viewTestGroup(elt, parentNode, i)
        else:
            pass

    def viewTestcase(self, modelDocument, parentNode, n):
        node = self.treeView.insert(parentNode, "end", modelDocument.objectId(),
                                    text=os.path.basename(modelDocument.uri),
                                    tags=("odd" if n & 1 else "even",))
        self.id += 1;
        if hasattr(modelDocument, "testcaseVariations"):
            for i, modelTestcaseVariation in enumerate(modelDocument.testcaseVariations):
                self.viewTestcaseVariation(modelTestcaseVariation, node, n + i + 1)

    def viewTestGroup(self, group, parentNode, n):
        node = self.treeView.insert(parentNode, "end", group.objectId(),
                                    text=group.get("name"),
                                    tags=("odd" if n & 1 else "even",))
        titleElt = XmlUtil.child(group, None, "title")
        if titleElt is not None:
            self.treeView.set(node, "name", titleElt.text)
        self.id += 1;
        i = -1
        for elt in group.iterchildren(tag="{http://www.w3.org/2005/02/query-test-XQTSCatalog}test-group"):
            i = i + 1
            self.viewTestGroup(elt, node, n + i + 1)
        for elt in group.iterchildren(tag="{http://www.w3.org/2005/02/query-test-XQTSCatalog}test-case"):
            if elt.get("is-XPath2") == "true":
                i = i + 1
                self.viewTestcaseVariation(elt, node, n + i + 1)

    def viewTestcaseVariation(self, modelTestcaseVariation, parentNode, n):
        if self.isTransformRegistry or modelTestcaseVariation.localName in ("testGroup", "test-case"):
            id = modelTestcaseVariation.name
        else:
            id = modelTestcaseVariation.id
            if id is None:
                id = ""
        node = self.treeView.insert(parentNode, "end", modelTestcaseVariation.objectId(),
                                    text=id,
                                    tags=("odd" if n & 1 else "even",))
        self.treeView.set(node, "name", (modelTestcaseVariation.description or modelTestcaseVariation.name))
        self.treeView.set(node, "readMeFirst", ",".join(str(uri) for uri in modelTestcaseVariation.readMeFirstUris))
        self.treeView.set(node, "status", modelTestcaseVariation.status)
        call = modelTestcaseVariation.cfcnCall
        if call: self.treeView.set(node, "call", call[0])
        test = modelTestcaseVariation.cfcnTest
        if test:
            self.treeView.set(node, "test", test[0])
        if getattr(self.modelXbrl.modelDocument, "outpath", None) and modelTestcaseVariation.resultIsInfoset:
            self.treeView.set(node, "infoset", modelTestcaseVariation.resultInfosetUri)
        _exp = sortCountExpected(modelTestcaseVariation.expected)
        self.treeView.set(node, "expected",
                          ", ".join(str(e) for e in _exp) if isinstance(_exp, list) else _exp)
        self.treeView.set(node, "actual", ", ".join(modelTestcaseVariation.actual))
        self.id += 1;

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            testcaseVariationId = modelObject.objectId()
            if self.treeView.exists(testcaseVariationId):
                if hasattr(modelObject, "status"):
                    self.treeView.set(testcaseVariationId, "status", modelObject.status)
                if hasattr(modelObject, "actual"):
                    self.treeView.set(testcaseVariationId, "actual", ", ".join(
                          str(code) for code in modelObject.actual))
                self.treeView.see(testcaseVariationId)
                self.treeView.selection_set(testcaseVariationId)
            self.blockViewModelObject -= 1



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import io
import json
import os
import subprocess
import sys
import threading
import tkinter.messagebox
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING
from urllib.error import URLError

import regex

from arelle import Version
from arelle.SystemInfo import getSystemInfo
from arelle.typing import TypeGetText

if TYPE_CHECKING:
    from arelle.CntlrWinMain import CntlrWinMain

_: TypeGetText

_MESSAGE_HEADER = "arelle\u2122 - Updater"
_SEMVER_PATTERN = regex.compile(
    r"(?P<semver>(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+))"
)
_UPDATE_URL = "https://api.github.com/repos/Arelle/Arelle/releases/latest"


@dataclass(eq=True, frozen=True, order=True)
class ArelleVersion:
    major: int
    minor: int
    patch: int

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


@dataclass(eq=True, frozen=True, order=True)
class ArelleRelease:
    version: ArelleVersion
    downloadUrl: str | None


def checkForUpdates(cntlr: CntlrWinMain) -> None:
    thread = threading.Thread(daemon=True, target=lambda c=cntlr: _checkForUpdates(c))
    thread.start()


def _checkForUpdates(cntlr: CntlrWinMain) -> None:
    if cntlr.webCache.workOffline:
        _showInfo(cntlr, _("Disable offline mode to check for updates."))
        return
    cntlr.showStatus(_("Checking for updates to Arelle"))
    try:
        arelleRelease = _getLatestArelleRelease(cntlr)
    except RuntimeError as e:
        _showWarning(
            cntlr,
            _("Failed to check for updates. Try again later. URL {0}, {1}").format(
                _UPDATE_URL, e
            ),
        )
        return
    finally:
        cntlr.showStatus("")  # clear web loading status entry
    cntlr.uiThreadQueue.put((_checkUpdateUrl, [cntlr, arelleRelease]))


def _getLatestArelleRelease(cntlr: CntlrWinMain) -> ArelleRelease:
    try:
        with io.BytesIO() as filestream:
            cntlr.webCache.retrieve(_UPDATE_URL, filestream=filestream)  # type: ignore[no-untyped-call]
            updateResponseJson: dict[str, Any] = json.load(filestream)
    except (URLError, json.JSONDecodeError) as e:
        raise RuntimeError("Failed to get latest Arelle release.") from e
    tagName = updateResponseJson.get("tag_name", "")
    try:
        updateVersion = _parseVersion(tagName)
    except ValueError as e:
        raise RuntimeError(f"Failed to parse version from latest Arelle release {tagName}.") from e
    downloadUrl = _getArelleReleaseDownloadUrl(updateResponseJson.get("assets", []))
    return ArelleRelease(version=updateVersion, downloadUrl=downloadUrl)


def _getArelleReleaseDownloadUrl(assets: list[dict[str, Any]]) -> str | None:
    if sys.platform == "darwin":
        if getSystemInfo().get('arch') == 'arm64':
            return _getArelleReleaseDownloadUrlByFileExtension(assets, ".dmg", nameIncludes='-arm64')
        return _getArelleReleaseDownloadUrlByFileExtension(assets, ".dmg", nameIncludes='-x64')
    elif sys.platform == "win32":
        return _getArelleReleaseDownloadUrlByFileExtension(assets, ".exe")
    else:
        return None


def _getArelleReleaseDownloadUrlByFileExtension(
    assets: list[dict[str, Any]], fileExtension: str, nameIncludes: str | None = None,
) -> str | None:
    for asset in assets:
        downloadUrl = asset.get("browser_download_url")
        if isinstance(downloadUrl, str) and downloadUrl.endswith(fileExtension):
            if nameIncludes is None or nameIncludes in downloadUrl:
                return downloadUrl
    return None


def _checkUpdateUrl(cntlr: CntlrWinMain, arelleRelease: ArelleRelease) -> None:
    try:
        currentVersion = _parseVersion(Version.version)
    except ValueError:
        _showWarning(cntlr, _("Unable to determine current version of Arelle."))
        return
    if arelleRelease.version > currentVersion:
        if arelleRelease.downloadUrl:
            reply = tkinter.messagebox.askokcancel(
                _(_MESSAGE_HEADER),
                _(
                    """
                    Update {0} is available, current version is {1}.

                    Download now?

                    (Arelle will exit before installing.)
                    """
                ).format(arelleRelease.version, currentVersion),
                parent=cntlr.parent,
            )
            if reply:
                _backgroundDownload(cntlr, arelleRelease)
        else:
            _showInfo(
                cntlr,
                _(
                    """
                    Update {0} is available, current version is {1}.

                    New version of Arelle can be downloaded at https://arelle.org/arelle/pub/.
                    """
                ).format(arelleRelease.version, currentVersion),
            )
    elif arelleRelease.version < currentVersion:
        _showInfo(
            cntlr,
            _("Current Arelle version {0} is newer than update {1}.").format(
                currentVersion, arelleRelease.version
            ),
        )
    else:
        _showInfo(
            cntlr,
            _("Arelle is already running the latest version {0}.").format(
                currentVersion
            ),
        )


def _parseVersion(versionStr: str) -> ArelleVersion:
    semverMatch = _SEMVER_PATTERN.search(versionStr)
    if semverMatch:
        return ArelleVersion(
            major=int(semverMatch.group("major")),
            minor=int(semverMatch.group("minor")),
            patch=int(semverMatch.group("patch")),
        )
    raise ValueError(f"Unable to parse version from {versionStr}")


def _backgroundDownload(cntlr: CntlrWinMain, arelleRelease: ArelleRelease) -> None:
    thread = threading.Thread(
        daemon=True, target=lambda: _download(cntlr, arelleRelease)
    )
    thread.start()


def _download(cntlr: CntlrWinMain, arelleRelease: ArelleRelease) -> None:
    if not arelleRelease.downloadUrl:
        raise RuntimeError(f"Arelle can't self-update on platform '{sys.platform}'.")
    filepathTmp = cntlr.webCache.getfilename(arelleRelease.downloadUrl, reload=True)
    if not filepathTmp:
        _showWarning(cntlr, _("Failed to download update."))
        return
    cntlr.showStatus(_("Download completed"), 5000)
    try:
        filepath = os.path.join(
            os.path.dirname(filepathTmp), os.path.basename(arelleRelease.downloadUrl)
        )
        os.rename(filepathTmp, filepath)
    except OSError:
        _showWarning(cntlr, _("Failed to process update."))
        return
    cntlr.uiThreadQueue.put((_install, [cntlr, filepath]))


def _install(cntlr: CntlrWinMain, filepath: str) -> None:
    if sys.platform == "win32":
        os.startfile(filepath)
    elif sys.platform == "darwin":
        try:
            subprocess.Popen(["open", filepath])
        except (OSError, subprocess.SubprocessError):
            _showWarning(cntlr, _("Failed to start updated Arelle instance."))
            return
    else:
        raise RuntimeError("Tried to install update on unsupported platform.")
    cntlr.uiThreadQueue.put((cntlr.quit, []))


def _showInfo(cntlr: CntlrWinMain, msg: str) -> None:
    tkinter.messagebox.showinfo(_(_MESSAGE_HEADER), msg, parent=cntlr.parent)


def _showWarning(cntlr: CntlrWinMain, msg: str) -> None:
    tkinter.messagebox.showwarning(_(_MESSAGE_HEADER), msg, parent=cntlr.parent)



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
import os
from arelle import ViewWinTree, ModelDtsObject, ModelInstanceObject, ModelRenderingObject, XbrlConst, XmlUtil, Locale
from arelle.ModelRelationshipSet import ModelRelationshipSet
from arelle.ModelDtsObject import ModelRelationship
from arelle.ModelFormulaObject import ModelFilter
from arelle.ModelObject import ModelObject
from arelle.ViewFileRelationshipSet import hasCalcArcrole
from arelle.ViewUtil import viewReferences, groupRelationshipSet, groupRelationshipLabel
from arelle.XbrlConst import conceptNameLabelRole, documentationLabel, widerNarrower

def viewRelationshipSet(modelXbrl, tabWin, arcrole,
                        linkrole=None, linkqname=None, arcqname=None, lang=None,
                        treeColHdr=None, showLinkroles=True, showRelationships=True, showColumns=True,
                        expandAll=False, hasTableIndex=False, noRelationshipsMsg=True):
    arcroleName = groupRelationshipLabel(arcrole)
    relationshipSet = groupRelationshipSet(modelXbrl, arcrole, linkrole, linkqname, arcqname)
    if not relationshipSet:
        if noRelationshipsMsg:
            modelXbrl.modelManager.addToLog(_("no relationships for {0}").format(arcroleName))
        return False
    modelXbrl.modelManager.showStatus(_("viewing relationships {0}").format(arcroleName))
    view = ViewRelationshipSet(modelXbrl, tabWin, arcrole, linkrole, linkqname, arcqname, lang,
                               treeColHdr, showLinkroles, showRelationships, showColumns,
                               expandAll, hasTableIndex)
    view.view(firstTime=True, relationshipSet=relationshipSet)
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

    # pop up menu
    menu = view.contextMenu()
    view.menuAddExpandCollapse()
    view.menuAddClipboard()
    view.menuAddLangs()
    view.menuAddLabelRoles(includeConceptName=True)
    view.menuAddViews()

    return True


class ViewRelationshipSet(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin,
                 arcrole, linkrole=None, linkqname=None, arcqname=None, lang=None,
                 treeColHdr=None, showLinkroles=True, showRelationships=True, showColumns=True,
                 expandAll=False, hasTableIndex=False):
        if isinstance(arcrole, (list,tuple)):
            tabName = arcrole[0]
        else:
            tabName = XbrlConst.baseSetArcroleLabel(arcrole)[1:]
        super(ViewRelationshipSet, self).__init__(modelXbrl, tabWin, tabName, True, lang)
        self.arcrole = arcrole
        self.linkrole = linkrole
        self.linkqname = linkqname
        self.arcqname = arcqname
        self.treeColHdr = treeColHdr
        self.showLinkroles = showLinkroles
        self.showRelationships = showRelationships
        self.showColumns = showColumns
        self.expandAllOnFirstDisplay = expandAll
        self.hasTableIndex = hasTableIndex
        self.isResourceArcrole = False

    def view(self, firstTime=False, relationshipSet=None):
        self.blockSelectEvent = 1
        self.blockViewModelObject = 0
        self.tag_has = defaultdict(list) # temporary until Tk 8.6
        # relationship set based on linkrole parameter, to determine applicable linkroles
        if relationshipSet is None:
            relationshipSet = groupRelationshipSet(self.modelXbrl, self.arcrole, self.linkrole, self.linkqname, self.arcqname)
        if not relationshipSet:
            self.modelXbrl.modelManager.addToLog(_("no relationships for {0}").format(groupRelationshipLabel(self.arcrole)))
            return False

        if firstTime:
            self.showReferences = False
            # set up treeView widget and tabbed pane
            hdr = self.treeColHdr if self.treeColHdr else _("{0} Relationships").format(groupRelationshipLabel(self.arcrole))
            self.treeView.heading("#0", text=hdr)
            if self.showColumns:
                if self.arcrole == XbrlConst.parentChild: # extra columns
                    self.treeView.column("#0", width=300, anchor="w")
                    self.treeView["columns"] = ("preferredLabel", "type", "references")
                    self.treeView.column("preferredLabel", width=64, anchor="w", stretch=False)
                    self.treeView.heading("preferredLabel", text=_("Pref. Label"))
                    self.treeView.column("type", width=100, anchor="w", stretch=False)
                    self.treeView.heading("type", text=_("Type"))
                    self.treeView.column("references", width=200, anchor="w", stretch=False)
                    self.treeView.heading("references", text=_("References"))
                elif hasCalcArcrole(self.arcrole):
                    self.treeView.column("#0", width=300, anchor="w")
                    self.treeView["columns"] = ("weight", "balance")
                    self.treeView.column("weight", width=48, anchor="w", stretch=False)
                    self.treeView.heading("weight", text=_("Weight"))
                    self.treeView.column("balance", width=70, anchor="w", stretch=False)
                    self.treeView.heading("balance", text=_("Balance"))
                elif self.arcrole == "XBRL-dimensions":    # add columns for dimensional information
                    self.treeView.column("#0", width=300, anchor="w")
                    self.treeView["columns"] = ("arcrole", "contextElement", "closed", "usable")
                    self.treeView.column("arcrole", width=100, anchor="w", stretch=False)
                    self.treeView.heading("arcrole", text="Arcrole")
                    self.treeView.column("contextElement", width=50, anchor="center", stretch=False)
                    self.treeView.heading("contextElement", text="Context")
                    self.treeView.column("closed", width=40, anchor="center", stretch=False)
                    self.treeView.heading("closed", text="Closed")
                    self.treeView.column("usable", width=40, anchor="center", stretch=False)
                    self.treeView.heading("usable", text="Usable")
                elif self.arcrole == "Table-rendering":    # add columns for dimensional information
                    self.treeView.column("#0", width=160, anchor="w")
                    self.treeView["columns"] = ("axis", "abstract", "merge", "header", "priItem", "dims")
                    self.treeView.column("axis", width=28, anchor="center", stretch=False)
                    self.treeView.heading("axis", text="Axis")
                    self.treeView.column("abstract", width=24, anchor="center", stretch=False)
                    self.treeView.heading("abstract", text="Abs")
                    self.treeView.column("merge", width=26, anchor="center", stretch=False)
                    self.treeView.heading("merge", text="Mrg")
                    self.treeView.column("header", width=160, anchor="w", stretch=False)
                    self.treeView.heading("header", text="Header")
                    self.treeView.column("priItem", width=100, anchor="w", stretch=False)
                    self.treeView.heading("priItem", text="Primary Item")
                    self.treeView.column("dims", width=150, anchor="w", stretch=False)
                    self.treeView.heading("dims", text=_("Dimensions"))
                elif self.arcrole == widerNarrower:
                    self.treeView.column("#0", width=300, anchor="w")
                    self.treeView["columns"] = ("wider", "documentation", "references")
                    self.treeView.column("wider", width=100, anchor="w", stretch=False)
                    self.treeView.heading("wider", text=_("Wider"))
                    self.treeView.column("documentation", width=200, anchor="w", stretch=False)
                    self.treeView.heading("documentation", text=_("Documentation"))
                    self.treeView.column("references", width=200, anchor="w", stretch=False)
                    self.treeView.heading("references", text=_("References"))
                elif isinstance(self.arcrole, (list,tuple)) or XbrlConst.isResourceArcrole(self.arcrole):
                    self.isResourceArcrole = True
                    self.showReferences = isinstance(self.arcrole, str) and self.arcrole.endswith("-reference")
                    self.treeView.column("#0", width=160, anchor="w")
                    self.treeView["columns"] = ("arcrole", "resource", "resourcerole", "lang")
                    self.treeView.column("arcrole", width=100, anchor="w", stretch=False)
                    self.treeView.heading("arcrole", text="Arcrole")
                    self.treeView.column("resource", width=60, anchor="w", stretch=False)
                    self.treeView.heading("resource", text="Resource")
                    self.treeView.column("resourcerole", width=100, anchor="w", stretch=False)
                    self.treeView.heading("resourcerole", text="Resource Role")
                    self.treeView.column("lang", width=36, anchor="w", stretch=False)
                    self.treeView.heading("lang", text="Lang")
        self.clearTreeView()
        self.id = 1

        # sort URIs by definition
        linkroleUris = []
        linkroleUriChildren = {}
        for linkroleUri in relationshipSet.linkRoleUris:
            modelRoleTypes = self.modelXbrl.roleTypes.get(linkroleUri)
            if modelRoleTypes:
                roledefinition = ((self.hasTableIndex and getattr(modelRoleTypes[0], "_tableIndex", False)) or
                                  self.modelXbrl.roleTypeDefinition(linkroleUri, self.lang))
                roleId = modelRoleTypes[0].objectId(self.id)
                if (self.hasTableIndex and hasattr(modelRoleTypes[0], "_tableChildren")):
                    linkroleUriChildren[linkroleUri] = [roleType.roleURI
                                                        for roleType in modelRoleTypes[0]._tableChildren]
            else:
                roledefinition = linkroleUri
                roleId = "node{0}".format(self.id)
            self.id += 1
            linkroleUris.append((roledefinition, linkroleUri, roleId))
        # entry may be ((table group, order, role definition), uri, id) or (str definition, uri, id)
        linkroleUris.sort(key=lambda d:d[0] if isinstance(d[0],tuple) else ("9noTableKey",d[0],""))

        def insertLinkroleChildren(parentNode, childUris):
            for childUri in childUris:
                for roledefinition, linkroleUri, roleId in linkroleUris:
                    if childUri == linkroleUri: # and isinstance(roledefinition, tuple): # tableGroup
                        if isinstance(roledefinition, tuple):
                            _nextTableGroup, _order, roledefinition = roledefinition
                        else:
                            modelRoleTypes = self.modelXbrl.roleTypes.get(linkroleUri)
                            roledefinition = (modelRoleTypes[0].genLabel(lang=self.lang, strip=True) or
                                              modelRoleTypes[0].definition or
                                              linkroleUri)
                        childId = "_{}{}".format(self.id, roleId)
                        self.id += 1
                        childNode = self.treeView.insert(parentNode, "end", childId, text=roledefinition, tags=("ELR",))
                        if childUri in linkroleUriChildren:
                            insertLinkroleChildren(childNode, linkroleUriChildren[childUri])
                        break

        # for each URI in definition order
        tableGroup = ""
        for roledefinition, linkroleUri, roleId in linkroleUris:
            if self.showLinkroles:
                if isinstance(roledefinition, tuple): # tableGroup
                    nextTableGroup, _order, roledefinition = roledefinition
                    if tableGroup != nextTableGroup:
                        self.treeView.insert("", "end", nextTableGroup, text=nextTableGroup[1:], tags=("Group",))
                        if not tableGroup: # first tableGroup item, expand it
                            self.setTreeItemOpen(nextTableGroup,open=True)
                        tableGroup = nextTableGroup
                linknode = self.treeView.insert(tableGroup, "end", roleId, text=roledefinition, tags=("ELR",))
                # add tableChildren as child nodes
                if linkroleUri in linkroleUriChildren:
                    insertLinkroleChildren(linknode, linkroleUriChildren[linkroleUri])
            else:
                linknode = ""
            if self.showRelationships:
                linkRelationshipSet = groupRelationshipSet(self.modelXbrl, self.arcrole, linkroleUri, self.linkqname, self.arcqname)
                for rootConcept in linkRelationshipSet.rootConcepts:
                    self.viewConcept(rootConcept, rootConcept, "", self.labelrole, linknode, 1, linkRelationshipSet, set())
                    self.tag_has[linkroleUri].append(linknode)

        if self.expandAllOnFirstDisplay:
            self.expandAll()

    def viewConcept(self, concept, modelObject, labelPrefix, preferredLabel, parentnode, n, relationshipSet, visited):
        if concept is None:
            return
        try:
            isRelation = isinstance(modelObject, ModelDtsObject.ModelRelationship)
            if isinstance(concept, ModelDtsObject.ModelConcept):
                text = labelPrefix + concept.label(preferredLabel,lang=self.lang,linkroleHint=relationshipSet.linkrole)
                if (self.arcrole in ("XBRL-dimensions", XbrlConst.hypercubeDimension) and
                    concept.isTypedDimension and
                    concept.typedDomainElement is not None):
                    text += " (typedDomain={0})".format(concept.typedDomainElement.qname)
            elif isinstance(concept, ModelInstanceObject.ModelFact):
                if concept.concept is not None:
                    text = labelPrefix + concept.concept.label(preferredLabel,lang=self.lang,linkroleHint=relationshipSet.linkrole)
                else:
                    text = str(concept.qname)
                if concept.contextID:
                    text += " [" + concept.contextID + "] = " + concept.effectiveValue
            elif self.arcrole == "Table-rendering":
                text = concept.localName
            elif isinstance(concept, ModelRenderingObject.ModelTable):
                text = (concept.genLabel(lang=self.lang, strip=True) or concept.localName)
            elif isinstance(concept, ModelDtsObject.ModelResource):
                if self.showReferences:
                    text = (concept.viewText() or concept.localName)
                else:
                    text = (Locale.rtlString(concept.textValue.strip(), lang=concept.xmlLang) or concept.localName)
            else:   # just a resource
                text = concept.localName
                # add recognized attributes
                if concept.localName == "enumeration" and concept.get("value"):
                    text += ' {}="{}"'.format("value", concept.get("value"))
            childnode = self.treeView.insert(parentnode, "end", modelObject.objectId(self.id), text=text, tags=("odd" if n & 1 else "even",))
            childRelationshipSet = relationshipSet
            if self.arcrole == XbrlConst.parentChild: # extra columns
                if isRelation:
                    preferredLabel = modelObject.preferredLabel
                    if preferredLabel and preferredLabel.startswith("http://"):
                        preferredLabel = os.path.basename(preferredLabel)
                    self.treeView.set(childnode, "preferredLabel", preferredLabel)
                self.treeView.set(childnode, "type", concept.niceType)
                self.treeView.set(childnode, "references", viewReferences(concept))
            elif hasCalcArcrole(self.arcrole):
                if isRelation:
                    self.treeView.set(childnode, "weight", "{:+0g} ".format(modelObject.weight))
                self.treeView.set(childnode, "balance", concept.balance)
            elif self.arcrole == "XBRL-dimensions" and isRelation: # extra columns
                relArcrole = modelObject.arcrole
                self.treeView.set(childnode, "arcrole", os.path.basename(relArcrole))
                if relArcrole in (XbrlConst.all, XbrlConst.notAll):
                    self.treeView.set(childnode, "contextElement", modelObject.contextElement)
                    self.treeView.set(childnode, "closed", modelObject.closed)
                elif relArcrole in (XbrlConst.dimensionDomain, XbrlConst.domainMember):
                    self.treeView.set(childnode, "usable", modelObject.usable)
                childRelationshipSet = self.modelXbrl.relationshipSet(XbrlConst.consecutiveArcrole.get(relArcrole,"XBRL-dimensions"),
                                                                      modelObject.consecutiveLinkrole)
            elif self.arcrole == "Table-rendering": # extra columns
                try:
                    header = concept.header(lang=self.lang,strip=True,evaluate=False)
                except AttributeError:
                    header = None # could be a filter
                if isRelation and header is None:
                    header = "{0} {1}".format(os.path.basename(modelObject.arcrole), concept.xlinkLabel)
                self.treeView.set(childnode, "header", header)
                if concept.get("abstract") == "true":
                    self.treeView.set(childnode, "abstract", '\u2713') # checkmark unicode character
                if concept.get("merge") == "true":
                    self.treeView.set(childnode, "merge", '\u2713') # checkmark unicode character
                if isRelation:
                    self.treeView.set(childnode, "axis", modelObject.axis)
                    if isinstance(concept, DefnMdlRuleDefinitionNode):
                        self.treeView.set(childnode, "priItem", concept.aspectValue(None, Aspect.CONCEPT))
                        self.treeView.set(childnode, "dims", ' '.join(("{0},{1}".format(dim, dimAspectVal.stringValue if isinstance(dimAspectVal,ModelObject) else
                                                                                        (dimAspectVal or concept.variableRefs()))
                                                                       for dim in (concept.aspectValue(None, Aspect.DIMENSIONS, inherit=False) or [])
                                                                       for dimAspectVal in (concept.aspectValue(None, dim),)
                                                                       if dimAspectVal is not None))
            )
            elif self.arcrole == widerNarrower:
                if isRelation:
                    otherWider = [modelRel.fromModelObject
                                  for modelRel in childRelationshipSet.toModelObject(concept)
                                  if modelRel.fromModelObject != modelObject.fromModelObject]
                    self.treeView.set(childnode, "wider", ", ".join(w.label(preferredLabel,lang=self.lang,linkroleHint=relationshipSet.linkrole) for w in otherWider))
                self.treeView.set(childnode, "documentation", concept.label(documentationLabel,lang=self.lang,linkroleHint=relationshipSet.linkrole,fallbackToQname=False))
                self.treeView.set(childnode, "references", viewReferences(concept))
            elif self.isResourceArcrole: # resource columns
                if isRelation:
                    self.treeView.set(childnode, "arcrole", os.path.basename(modelObject.arcrole))
                if isinstance(concept, ModelDtsObject.ModelResource):
                    self.treeView.set(childnode, "resource", concept.localName)
                    self.treeView.set(childnode, "resourcerole", os.path.basename(concept.role or ''))
                    self.treeView.set(childnode, "lang", concept.xmlLang)
            self.id += 1
            self.tag_has[modelObject.objectId()].append(childnode)
            if isRelation:
                self.tag_has[modelObject.toModelObject.objectId()].append(childnode)
            if concept not in visited:
                visited.add(concept)
                for modelRel in childRelationshipSet.fromModelObject(concept):
                    nestedRelationshipSet = childRelationshipSet
                    targetRole = modelRel.targetRole
                    if hasCalcArcrole(self.arcrole):
                        childPrefix = "({:0g}) ".format(modelRel.weight) # format without .0 on integer weights
                    elif targetRole is None or len(targetRole) == 0:
                        targetRole = relationshipSet.linkrole
                        childPrefix = ""
                    else:
                        nestedRelationshipSet = self.modelXbrl.relationshipSet(childRelationshipSet.arcrole, targetRole)
                        childPrefix = "(via targetRole) "
                    toConcept = modelRel.toModelObject
                    if toConcept in visited:
                        childPrefix += "(loop)"
                    labelrole = modelRel.preferredLabel
                    if not labelrole or self.labelrole == conceptNameLabelRole:
                        labelrole = self.labelrole
                    n += 1 # child has opposite row style of parent
                    self.viewConcept(toConcept, modelRel, childPrefix, labelrole, childnode, n, nestedRelationshipSet, visited)
                visited.remove(concept)
        except AttributeError:
            return # bad object, don't try to display

    def getToolTip(self, tvRowId, tvColId):
        # override tool tip when appropriate
        if self.arcrole == "Table-rendering" and tvColId in ("#0", "#4"):
            try:
                modelObject = self.modelXbrl.modelObject(tvRowId) # this is a relationship object
                if isinstance(modelObject, ModelRelationship):
                    modelResource = modelObject.toModelObject
                else:
                    modelResource = modelObject
                if tvColId == "#0":
                    return modelResource.definitionNodeView
                elif tvColId == "#4":
                    return " \n".join(": ".join(l)
                                      for l in modelResource.definitionLabelsView)
                else:
                    return None
            except (AttributeError, KeyError):
                try: # rendering filter relationships
                    if tvColId == "#0":
                        modelResource = modelObject.toModelObject
                        return "{0}: {1}\ncomplement: {2}\ncover: {3}\n{4}".format(
                                modelResource.localName, modelResource.viewExpression,
                                str(modelObject.isComplemented).lower(),
                                str(modelObject.isCovered).lower(),
                                '\n'.join("{0}: {1}".format(p[0], p[1])
                                          for p in modelResource.propertyView
                                          if p and len(p) >= 2))
                except (AttributeError, KeyError):
                    pass
        return None

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            selection = self.treeView.selection()
            if selection is not None and len(selection)>0:
                self.modelXbrl.viewModelObject(selection[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                # check if modelObject is a relationship in given linkrole
                if isinstance(modelObject, ModelDtsObject.ModelRelationship):
                    linkroleId = self.tag_has.get(modelObject.linkrole)
                    if linkroleId:
                        linkroleId = linkroleId[0]
                    else:
                        linkroleId = None
                else:
                    linkroleId = None
                # get concept of fact or toConcept of relationship, role obj if roleType
                conceptId = modelObject.viewConcept.objectId()
                items = self.tag_has.get(conceptId)
                if items:
                    for item in items:
                        if self.treeView.exists(item):
                            if linkroleId is None or self.hasAncestor(item, linkroleId):
                                self.treeView.see(item)
                                self.treeView.selection_set(item)
                                break
            except (AttributeError, KeyError):
                    self.treeView.selection_set(())
            self.blockViewModelObject -= 1

    def hasAncestor(self, node, ancestor):
        if node == ancestor:
            return True
        elif node:
            return self.hasAncestor(self.treeView.parent(node), ancestor)
        return False

from arelle.ModelRenderingObject import DefnMdlRuleDefinitionNode
from arelle.ModelFormulaObject import Aspect



"""
See COPYRIGHT.md for copyright information.
"""
import regex as re
from arelle.ModelValue import qname
from arelle.XbrlConst import xsd

# MERGE TO arelle.XbrlConst when promoting plugin to infrastructure

oimTaxonomyDocTypePattern = re.compile(r"\s*\{.*\"documentType\"\s*:\s*\"https://xbrl.org/[0-9]{4}/taxonomy\"", flags=re.DOTALL)
oimTaxonomyDocTypes = (
        "https://xbrl.org/2025/taxonomy",
    )

xbrl = "https://xbrl.org/2025"

qnStdLabel = qname(xbrl, "xbrli:label")
qnXsDate = qname(xsd, "xs:date")
qnXsDateTime = qname(xsd, "xs:dateTime")
qnXsDuration = qname(xsd, "xs:duration")
qnXsQName = qname(xsd, "xs:QName")

qnXbrlLabelObj = qname(xbrl, "xbrl:labelObject")

objectsWithProperties = {
    qname(xbrl, "xbrl:taxonomyObject"),
    qname(xbrl, "xbrl:conceptObject"),
    qname(xbrl, "xbrl:abstractObject"),
    qname(xbrl, "xbrl:cubeObject"),
    qname(xbrl, "xbrl:dimensionObject"),
    qname(xbrl, "xbrl:domainObject"),
    qname(xbrl, "xbrl:entityObject"),
    qname(xbrl, "xbrl:groupObject"),
    qname(xbrl, "xbrl:networkObject"),
    qnXbrlLabelObj,
    qname(xbrl, "xbrl:memberObject"),
    qname(xbrl, "xbrl:referenceObject"),
    qname(xbrl, "xbrl:relationshipObject"),
    }

xbrlTaxonomyObjects = {
    "documentInfo": {
        "documentType": oimTaxonomyDocTypes[0],
        "namespaces": {
            "xbrl": xbrl,
            "xbrli": "https://xbrl.org/2025/instance",
            "xs": xsd
        }
    },
    "taxonomy": {
        "name": "xbrl:BuiltInCoreObjectsTaxonomy",
        "frameworkName": "types",
        "version": "2025",
        "resolved": False,
        "dataTypes": [
            {
                "name": "xs:string",
                "baseType": "xs:string"
            },
            {
                "name": "xs:boolean",
                "baseType": "xs:boolean"
            },
            {
                "name": "xs:decimal",
                "baseType": "xs:decimal"
            },
            {
                "name": "xs:float",
                "baseType": "xs:float"
            },
            {
                "name": "xs:double",
                "baseType": "xs:double"
            },
            {
                "name": "xs:duration",
                "baseType": "xs:duration"
            },
            {
                "name": "xs:dateTime",
                "baseType": "xs:dateTime"
            },
            {
                "name": "xs:time",
                "baseType": "xs:time"
            },
            {
                "name": "xs:date",
                "baseType": "xs:date"
            },
            {
                "name": "xs:gYearMonth",
                "baseType": "xs:gYearMonth"
            },
            {
                "name": "xs:gYear",
                "baseType": "xs:gYear"
            },
            {
                "name": "xs:gMonthDay",
                "baseType": "xs:gMonthDay"
            },
            {
                "name": "xs:gDay",
                "baseType": "xs:gDay"
            },
            {
                "name": "xs:gMonth",
                "baseType": "xs:gMonth"
            },
            {
                "name": "xs:hexBinary",
                "baseType": "xs:hexBinary"
            },
            {
                "name": "xs:base64Binary",
                "baseType": "xs:base64Binary"
            },
            {
                "name": "xs:anyURI",
                "baseType": "xs:anyURI"
            },
            {
                "name": "xs:QName",
                "baseType": "xs:QName"
            },
            {
                "name": "xs:NOTATION",
                "baseType": "xs:NOTATION"
            },
            {
                "name": "xs:normalizedString",
                "baseType": "xs:string"
            },
            {
                "name": "xs:token",
                "baseType": "xs:normalizedString"
            },
            {
                "name": "xs:language",
                "baseType": "xs:token"
            },
            {
                "name": "xs:NMTOKEN",
                "baseType": "xs:token"
            },
            {
                "name": "xs:NMTOKENS",
                "baseType": "xs:NMTOKEN"
            },
            {
                "name": "xs:Name",
                "baseType": "xs:token"
            },
            {
                "name": "xs:NCName",
                "baseType": "xs:Name"
            },
            {
                "name": "xs:ID",
                "baseType": "xs:NCName"
            },
            {
                "name": "xs:IDREF",
                "baseType": "xs:NCName"
            },
            {
                "name": "xs:IDREFS",
                "baseType": "xs:IDREF"
            },
            {
                "name": "xs:ENTITY",
                "baseType": "xs:IDREF"
            },
            {
                "name": "xs:ENTITIES",
                "baseType": "xs:ENTITY"
            },
            {
                "name": "xs:integer",
                "baseType": "xs:decimal"
            },
            {
                "name": "xs:nonPositiveInteger",
                "baseType": "xs:integer",
                "maxInclusive": 0

            },
            {
                "name": "xs:negativeInteger",
                "baseType": "xs:nonPositiveInteger",
                "maxInclusive": -1
            },
            {
                "name": "xs:long",
                "baseType": "xs:integer"
            },
            {
                "name": "xs:int",
                "baseType": "xs:long",
                "maxInclusive": 2147483647,
                "minInclusive":  -2147483648
            },
            {
                "name": "xs:short",
                "baseType": "xs:int",
                "maxInclusive": 32767,
                "minInclusive":  -32768
            },
            {
                "name": "xs:byte",
                "baseType": "xs:short",
                "maxInclusive": 127,
                "minInclusive":  -128
            },
            {
                "name": "xs:nonNegativeInteger",
                "baseType": "xs:integer",
                "minInclusive":  0
            },
            {
                "name": "xs:unsignedLong",
                "baseType": "xs:nonNegativeInteger",
                "maxInclusive": 18446744073709551615
            },
            {
                "name": "xs:unsignedInt",
                "baseType": "xs:unsignedInt",
                "maxInclusive": 4294967295
            },
            {
                "name": "xs:unsignedShort",
                "baseType": "xs:unsignedInt",
                "maxInclusive": 65535
            },
            {
                "name": "xs:unsignedByte",
                "baseType": "xs:unsignedShort",
                "maxInclusive": 255
            },
            {
                "name": "xs:positiveInteger",
                "baseType": "xs:nonNegativeInteger",
                "minInclusive":  1
            },
            {
                "name": "xs:yearMonthDuration",
                "baseType": "xs:duration"
            },
            {
                "name": "xs:dayTimeDuration",
                "baseType": "xs:duration"
            },
            {
                "name": "xs:dateTimeStamp",
                "baseType": "xs:dateTime"
            }
        ]
    }
}



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zipfile

    from arelle.FileSource import FileSource


def isZipfileEncrypted(info: zipfile.ZipInfo) -> bool:
    return bool(info.flag_bits & 0x1)

def getPackageTopLevelFiles(filesource: FileSource) -> set[str]:
    return {e for e in filesource.dir or [] if "/" not in e}

def getPackageTopLevelDirectories(filesource: FileSource) -> set[str]:
    return {e.partition("/")[0] for e in filesource.dir or [] if "/" in e}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations


class ElementBase:
    def __init__(self, *children: ElementBase, attrib: dict[str, str] | None = None, nsmap: dict[str, str] | None = None, **extra: str) -> None:
        self._children = children
        self._attrib = attrib if attrib is not None else {}
        self._nsmap = nsmap if nsmap is not None else {}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

from .ReportFolderType import ReportFolderType


@dataclass(frozen=True)
class UploadContents:
    reports: dict[ReportFolderType, frozenset[Path]]
    uploadPaths: list[UploadPathInfo]

    @property
    def sortedPaths(self) -> list[Path]:
        return sorted(uploadPath.path for uploadPath in self.uploadPaths)

    @cached_property
    def uploadPathsByFullPath(self) -> dict[Path, UploadPathInfo]:
        return {
            uploadPath.fullPath: uploadPath
            for uploadPath in self.uploadPaths
        }

    @cached_property
    def uploadPathsByPath(self) -> dict[Path, UploadPathInfo]:
        return {
            uploadPath.path: uploadPath
            for uploadPath in self.uploadPaths
        }


@dataclass(frozen=True)
class UploadPathInfo:
    fullPath: Path
    isAttachment: bool
    isCorrection: bool
    isCoverPage: bool
    isDirectory: bool
    isRoot: bool
    isSubdirectory: bool
    path: Path
    reportFolderType: ReportFolderType | None
    reportPath: Path | None



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Iterable

from collections import defaultdict
from jaconv import jaconv

from arelle.ModelDocument import ModelDocument
from arelle.ModelObject import ModelObject
from arelle.typing import TypeGetText
from arelle.utils.validate.Validation import Validation

_: TypeGetText

# Table of content number sets per EDINET documentation:
# Figure "3-4-5 設定可能な目次番号の一覧" in "File Specification for EDINET Filing"
# https://disclosure2dl.edinet-fsa.go.jp/guide/static/disclosure/download/ESE140104.pdf
TOC_DIGITS = [
    '一',
    '二',
    '三',
    '四',
    '五',
    '六',
    '七',
    '八',
    '九',
    '十',
]
FULL_WIDTH_DIGIT_MAP = {
    str(d): jaconv.h2z(str(d), kana=True, ascii=True, digit=True)
    for d in range(0, 10)
}
KATAKANA_GOJUON_SEQUENCE = [
    # a-column
    'ア', 'イ', 'ウ', 'エ', 'オ',
    # k-column
    'カ', 'キ', 'ク', 'ケ', 'コ',
    # s-column
    'サ', 'シ', 'ス', 'セ', 'ソ',
    # t-column
    'タ', 'チ', 'ツ', 'テ', 'ト',
    # n-column
    'ナ', 'ニ', 'ヌ', 'ネ', 'ノ',
    # h-column
    'ハ', 'ヒ', 'フ', 'ヘ', 'ホ',
    # m-column
    'マ', 'ミ', 'ム', 'メ', 'モ',
    # y-column
    'ヤ', 'ユ', 'ヨ',
    # r-column
    'ラ', 'リ', 'ル', 'レ', 'ロ',
    # w-column
    'ワ', 'ヲ',
    # n-row
    'ン'
]
TOC_NUMBER_SETS = {
    1: {
        '第一部': [f'第{d}部' for d in TOC_DIGITS] + [f'第十{d}部' for d in TOC_DIGITS[:-1]] + ['第二十部']
    },
    2: {
        '第１': [
            '第' + ''.join(FULL_WIDTH_DIGIT_MAP[ddd] for ddd in dd)
            for dd in [
                str(d) for d in range(1, 61)
            ]
        ],
    },
    3: {
        '１': [
            ''.join(FULL_WIDTH_DIGIT_MAP[ddd] for ddd in dd)
            for dd in [
                str(d) for d in range(1, 61)
            ]
        ],
    },
    4: {
        '（１）': [
            '（' + ''.join(FULL_WIDTH_DIGIT_MAP[ddd] for ddd in dd) + '）'
            for dd in [
                str(d) for d in range(1, 61)
            ]
        ],
    },
    5: {
        '①': [chr(ord('①') + d) for d in range(0, 20)],
        f'（{KATAKANA_GOJUON_SEQUENCE[0]}）': [f'（{d}）' for d in KATAKANA_GOJUON_SEQUENCE],
        KATAKANA_GOJUON_SEQUENCE[0]: KATAKANA_GOJUON_SEQUENCE,
        '（ａ）': ['（' + chr(ord('ａ') + d) + '）' for d in range(0, 26)],
        'ａ': [chr(ord('ａ') + d) for d in range(0, 26)],
    },
}
SHALLOWEST_LEVEL = min(TOC_NUMBER_SETS.keys())
DEEPEST_LEVEL = max(TOC_NUMBER_SETS.keys())


class TableOfContentsBuilder:
    _currentDocument: ModelDocument | None
    _currentLevel: int
    _documents: list[ModelDocument]
    _floatingLevel: int | None
    _levelLabels: dict[int, set[str]]
    _levelPositions: dict[int, int]
    _levelSequences: dict[int, list[str] | None]
    _tocEntryCount: int
    _tocSequence: list[tuple[str, str, ModelObject]]
    _validations: list[Validation]

    def __init__(self) -> None:
        self._currentDocument = None
        self._currentLevel = SHALLOWEST_LEVEL
        self._floatingLevel = None
        self._levelLabels = defaultdict(set)
        self._levelPositions = defaultdict(int)
        self._levelSequences = {}
        self._documents = []
        self._tocEntryCount = 0
        self._tocSequence = []
        self._validations = []

    def _build(self) -> None:
        documents = sorted(self._documents, key=lambda doc: doc.basename)
        for document in documents:
            rootElt = document.xmlRootElement
            for elt in rootElt.iterdescendants():
                if not isinstance(elt, ModelObject):
                    continue
                if elt.elementQname.localName == 'title':
                    continue
                if elt.text is not None:
                    self._element(elt)

    def _checkCurrentLevelNext(self, number: str, nextPosition: int, currentSequence: list[str] | None) -> bool:

        # NEXT IN SEQUENCE
        # We can only move ahead in the sequence if it has been established AND
        # we have not reached the end of the allowed numbers in the sequence.
        if currentSequence is None:
            # No sequence established at this level.
            return False
        if nextPosition >= len(currentSequence):
            # We have reached the end of the sequence at this level.
            return False
        nextNumberInSequence = currentSequence[nextPosition]
        if number == nextNumberInSequence:
            # Increment the position at the current level.
            self._levelPositions[self._currentLevel] = self._levelPositions[self._currentLevel] + 1
            # Reset the floating status.
            self._floatingLevel = None
            return True
        return False

    def _checkDeepLevel(self, number: str) -> bool:
        # STARTING DEEPER SEQUENCE
        # We can move deeper one or more levels.
        if self._floatingLevel is not None:
            # We don't move deeper if we are floating.
            return False
        deeperStartingNumbers = {
            startingNumber: level
            for level in range(min(self._currentLevel + 1, DEEPEST_LEVEL), DEEPEST_LEVEL + 1)
            for startingNumber in TOC_NUMBER_SETS[level]
        }
        if number not in deeperStartingNumbers:
            # The number is not the first entry in a deeper level sequence.
            return False
        # Set the level to the deeper level.
        self._currentLevel = deeperStartingNumbers[number]
        # Establish the sequence at the deeper level.
        self._levelSequences[self._currentLevel] = TOC_NUMBER_SETS[self._currentLevel][number]
        # Reset the duplicates at the deeper level.
        self._levelLabels[self._currentLevel].clear()
        # Reset the position at the deeper level.
        self._levelPositions[self._currentLevel] = 1
        return True

    def _checkShallowLevel(self, number: str) -> bool:
        # RESUMING SHALLOWER SEQUENCE
        # We may be moving back up one or more levels.
        nextLevel = None
        for shallowLevel in reversed(range(SHALLOWEST_LEVEL, self._currentLevel)):
            shallowSequence = self._levelSequences.get(shallowLevel)
            # For each level, should we be checking earlier and later numbers in the sequence
            # to fire EC3002E or EC3003E?
            if shallowSequence is not None and number == shallowSequence[self._levelPositions[shallowLevel]]:
                # This is the next number in the upper sequence.
                # Move up one level.
                nextLevel = shallowLevel
                break
            shallowLevel -= 1
        if nextLevel is None:
            return False
        # For each level shallower than the new level down to the current level...
        for i in range(nextLevel + 1, self._currentLevel + 1):
            # Reset the position.
            self._levelPositions[i] = 0
            # Reset the duplicates.
            self._levelLabels[i].clear()
            # Reset the sequence.
            self._levelSequences[i] = None
        # Set the level to the shallower level.
        self._currentLevel = nextLevel
        # Increment the position at the shallower level.
        self._levelPositions[self._currentLevel] = self._levelPositions[self._currentLevel] + 1
        # Reset the floating status.
        self._floatingLevel = None
        return True

    def _closeDocument(self) -> None:
        assert self._currentDocument is not None, "No document is currently open."
        # EDINET.EC2001E: There must be at least one table of contents entry in each file.
        if self._tocEntryCount == 0:
            self._validations.append(Validation.error(
                codes='EDINET.EC2001E',
                msg=_("The table of contents is not listed at the beginning. "
                      "File name: '%(path)s'. "
                      "Please provide the table of contents entry for the file."),
                path=self._currentDocument.basename,
            ))
        self._currentDocument = None

    def _element(self, elt: ModelObject) -> None:
        # New document, close previous.
        if self._currentDocument is not None and self._currentDocument != elt.document:
            self._closeDocument()
        # First or new document, open document.
        if self._currentDocument is None:
            self._openDocument(elt.document)
        text = elt.text or ""
        openIndex = -1
        closeIndex = -1
        if "【" in text:
            openIndex = text.index("【") + 1
            # EDINET.EC2004E: Corner brackets (【) must not overlap.
            if "【" in text[openIndex:]:
                self._validations.append(Validation.error(
                    codes='EDINET.EC2004E',
                    msg=_("The opening bracket (【) is repeated. "
                          "File name: '%(path)s' (line %(line)s). "
                          "When viewed in a browser, it is not possible to display "
                          "more than one table of contents item on one line. Please "
                          "delete the brackets (【) in the relevant file."),
                    path=elt.document.basename,
                    line=elt.sourceline,
                    modelObject=elt,
                ))
        if "】" in text:
            closeIndex = text.index("】") - 1
        if openIndex >= 0:
            if closeIndex > openIndex:
                number = text[:openIndex - 1].strip()
                label = text[openIndex:closeIndex + 1]
                self._tocSequence.append((number, label, elt))
                self._tocEntryCount += 1
            else:
                # EDINET.EC2007E: The table of contents entries must be enclosed in square brackets (】).
                self._validations.append(Validation.error(
                    codes='EDINET.EC2007E',
                    msg=_("The table of contents entry is not closed with '】'. "
                          "File name: '%(path)s' (line %(line)s). "
                          "Add a closing bracket (】) to match the open bracket (【) "
                          "in the table of contents entry for the file in question."),
                    path=elt.document.basename,
                    line=elt.sourceline,
                    modelObject=elt,
                ))

    def _isFloating(self) -> bool:
        return self._floatingLevel is not None or self._currentLevel >= DEEPEST_LEVEL

    def _normalizeNumber(self, number: str) -> str:
        # EDINET does not support:
        # - Mixture of half-width and full-width digits within a number.
        # - Mixture of half-width and full-width parentheses within a number.
        # EDINET does support:
        # - Mixture of half-width digits with full-width parentheses, and vice versa.
        # We will normalize to full-width parantheses and digits for number validation.
        if "(" in number and "）" in number: # Half-width (, full-width ）
            return number
        if "（" in number and ")" in number: # Full-width (, half-width )
            return number
        paranthesesFullWidth: bool | None = None
        numbersFullWidth: bool | None = None
        for c in number:
            if c in ("(", ")"):
                if paranthesesFullWidth == True:
                    return number # Mix of half/full-width parantheses
                paranthesesFullWidth = False
            elif c in ("（", " ）"):
                if paranthesesFullWidth == False:
                    return number # Mix of half/full-width parantheses
                paranthesesFullWidth = True
            elif c in FULL_WIDTH_DIGIT_MAP:
                if numbersFullWidth == True:
                    return number # Mix of half/full-width digits
                numbersFullWidth = False
            elif c in FULL_WIDTH_DIGIT_MAP.values():
                if numbersFullWidth == False:
                    return number # Mix of half/full-width digits
                numbersFullWidth = True
        return jaconv.h2z(number, kana=True, ascii=True, digit=True)

    def _openDocument(self, modelDocument: ModelDocument) -> None:
        assert self._currentDocument is None, "Close current document before opening another."
        self._tocEntryCount = 0
        self._currentDocument = modelDocument

    def _validateItem(self, number: str, label: str, elt: ModelObject) -> Iterable[Validation]:
        # Convert to full-width, ONLY if fully half-width.
        # EDINET does not support a mixture of half-width and full-width digits in TOC numbers.
        number = self._normalizeNumber(number)
        nextPosition = self._levelPositions[self._currentLevel]
        currentSequence = self._levelSequences.get(self._currentLevel)

        # UN-NUMBERED (FLOATING) ITEM
        # Floating items trigger special behavior for following items
        # until the current or shallower level is resumed.
        if number == "":
            # Only trigger floating/warning if we are not already floating, and we
            # aren't already at the deepest level.
            if not self._isFloating():
                # EDINET.EC2002W: The table of contents number must be present.
                # Note from documentation: Even if the data content is normal, it may be identified as an
                # exception and a warning may be displayed.
                yield Validation.warning(
                    codes='EDINET.EC2002W',
                    msg=_("The table of contents number is not listed. "
                          "File name: '%(path)s' (line %(line)s). "
                          "Please include the table of contents number of the relevant file."),
                    path=elt.document.basename,
                    line=elt.sourceline,
                    modelObject=elt,
                )
                self._floatingLevel = self._currentLevel
            return

        if self._checkCurrentLevelNext(number, nextPosition, currentSequence):
            return
        if self._checkDeepLevel(number):
            return
        if self._checkShallowLevel(number):
            return

        # OTHER NUMBER IN CURRENT SEQUENCE
        if currentSequence is not None and  number in currentSequence:
            numberIndex = currentSequence.index(number)
            assert numberIndex != nextPosition
            # Is it repeating a previous number?
            if numberIndex < nextPosition:
                # EDINET.EC3002E: Table of contents numbers for table of contents entries
                # must not be repeated within the same hierarchy.
                yield Validation.error(
                    codes='EDINET.EC3002E',
                    msg=_("The table of contents number of the table of contents "
                          "item is duplicated in the same hierarchy. "
                          "File name: '%(path)s' (line %(line)s). "
                          "Please correct the table of contents number in the "
                          "table of contents entry of the relevant file."),
                    path=elt.document.basename,
                    line=elt.sourceline,
                    modelObject=elt,
                )
                return
            # Is it skipping ahead?
            if numberIndex > nextPosition:
                # EDINET.EC3003E: There must be no gaps in the table of contents numbers
                # within the same hierarchy.
                yield Validation.error(
                    codes='EDINET.EC3003E',
                    msg=_("There is a gap in the table of contents number "
                          "within the same hierarchy. "
                          "File name: '%(path)s' (line %(line)s). "
                          "Please enter the missing table of contents number "
                          "in the appropriate file."),
                    number=number,
                    label=label,
                    path=elt.document.basename,
                    line=elt.sourceline,
                    modelObject=elt,
                )
                # Jump ahead to minimize further errors.
                # Increment the position at the current level.
                self._levelPositions[self._currentLevel] = numberIndex + 1
                # Reset the floating status.
                self._floatingLevel = None
                return

        # INVALID NUMBER
        # Not un-numbered (floating), not next in sequence, not starting deeper sequence,
        # not resuming shallower sequence.
        if self._floatingLevel is None:
            # The difference between EC3004W and EC3005E is unclear based on documentation.
            # We will implement the higher level severity version of the two.
            # EDINET.EC3004W: The table of contents number of the table of contents
            # item must be set.
            # EDINET.EC3005E: The table of contents numbers for the table of contents
            # entries must be as specified in the format.
            yield Validation.error(
                codes='EDINET.EC3005E',
                msg=_("The table of contents number for item '%(number)s %(label)s' is incorrect. "
                      "File name: '%(path)s' (line %(line)s). "
                      "Please correct the table of contents number of the "
                      "table of contents item of the corresponding file."),
                number=number,
                label=label,
                path=elt.document.basename,
                line=elt.sourceline,
                modelObject=elt,
            )

    def addDocument(self, modelDocument: ModelDocument) -> None:
        self._documents.append(modelDocument)

    def validate(self) -> Iterable[Validation]:
        self._build()
        if self._currentDocument is not None:
            self._closeDocument()
        # Yield errors encountered during loading/build.
        yield from self._validations

        # Tracks the current level.
        self._currentLevel = SHALLOWEST_LEVEL
        # Tracks the current position in the sequence at each level.
        self._levelPositions: dict[int, int] = defaultdict(int)
        # Tracks the active number set at each level.
        self._levelSequences: dict[int, list[str] | None] = {
            self._currentLevel: next(iter(TOC_NUMBER_SETS[self._currentLevel].values()))
        }
        # Tracks unique labels within a sequence.
        self._levelLabels = defaultdict(set)
        # Tracks floating status.
        self._floatingLevel = None
        for number, label, elt in self._tocSequence:
            yield from self._validateItem(number, label, elt)

            # We are only concerned about duplicates if we are not floating.
            if not self._isFloating():
                if label in self._levelLabels[self._currentLevel]:
                    # EDINET.EC2005E: Table of contents entries must not be duplicated.
                    # Note: Sample filings suggest this applies to entries that are
                    # siblings within the hierarchy.
                    yield Validation.error(
                        codes='EDINET.EC2005E',
                        msg=_("The table of contents item ('%(label)s') is duplicated. "
                              "File name: '%(path)s' (line %(line)s). "
                              "Please remove the duplicate table of contents in the appropriate file."),
                        label=label,
                        path=elt.document.basename,
                        line=elt.sourceline,
                        modelObject=elt,
                    )
                else:
                    self._levelLabels[self._currentLevel].add(label)

            if not self._isFloating():
                # EDINET.EC2003E: The table of contents must be no longer than 384 bytes
                # (equivalent to 128 full-width characters).
                b = label.encode('utf-8')
                if len(label.encode('utf-8')) > 384:
                    yield Validation.error(
                        codes='EDINET.EC2003E',
                        msg=_("The table of contents entry exceeds 384 bytes. "
                              "File name: '%(path)s' (line %(line)s). "
                              "Please modify the table of contents of the relevant "
                              "file so that it is within 384B (bytes) (equivalent to "
                              "128 full-width characters)."),
                        path=elt.document.basename,
                        line=elt.sourceline,
                        modelObject=elt,
                    )

                # EDINET.EC2008E: Table of contents entries must not contain HTML tags.
                if len(elt) > 0:
                    yield Validation.error(
                        codes='EDINET.EC2008E',
                        msg=_("The table of contents contains HTML tags. "
                              "File name: '%(path)s' (line %(line)s). "
                              "HTML tags are not allowed in table of contents entries. "
                              "Please remove the HTML tags from the relevant file."),
                        path=elt.document.basename,
                        line=elt.sourceline,
                        modelObject=elt,
                    )

            # Uncomment for debugging output of TOC structure.
            # print(
            #     f'{" " if self._floatingLevel is not None else "|"}\t' * (self._currentLevel - 1) +
            #     f"{number} [{label}] \t{elt.document.basename}"
            # )



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
from arelle import ViewWinTree, ModelDtsObject, XbrlConst, XmlUtil, Locale

def viewRoleTypes(modelXbrl, tabWin, isArcrole=False, lang=None):
    modelXbrl.modelManager.showStatus(_("viewing arcrole types") if isArcrole else _("viewing role types"))
    view = ViewRoleTypes(modelXbrl,
                         tabWin,
                         "Arcrole Types" if isArcrole else "Role Types",
                         isArcrole,
                         lang)
    view.view(firstTime=True)
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

    # pop up menu
    menu = view.contextMenu()
    view.menuAddClipboard()
    view.menuAddLangs()
    view.menuAddLabelRoles(includeConceptName=True)
    view.menuAddViews()


class ViewRoleTypes(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin, header, isArcrole, lang=None):
        super(ViewRoleTypes, self).__init__(modelXbrl, tabWin, header, True, lang)
        self.isArcrole = isArcrole

    def view(self, firstTime=False):
        self.blockSelectEvent = 1
        self.blockViewModelObject = 0
        self.tag_has = defaultdict(list) # temporary until Tk 8.6

        roletypes = self.modelXbrl.arcroleTypes if self.isArcrole else self.modelXbrl.roleTypes
        if not roletypes:
            return False # nothing to display

        if firstTime:
            # set up treeView widget and tabbed pane
            hdr = _("Arcrole Types") if self.isArcrole else _("Role Types")
            self.treeView.heading("#0", text=hdr)
            self.treeView.column("#0", width=300, anchor="w")
            if self.isArcrole:
                self.treeView["columns"] = ("definition", "cyclesAllowed", "usedOn")
            else:
                self.treeView["columns"] = ("definition", "usedOn")
            self.treeView.column("definition", width=300, anchor="w", stretch=True)
            self.treeView.heading("definition", text=_("Definition"))
            if self.isArcrole:
                self.treeView.column("cyclesAllowed", width=60, anchor="w", stretch=False)
                self.treeView.heading("cyclesAllowed", text=_("Cycles Allowed"))
            self.treeView.column("usedOn", width=200, anchor="w", stretch=False)
            self.treeView.heading("usedOn", text=_("Used On"))
            self.setColumnsSortable()
        self.id = 1
        self.clearTreeView()
        # sort URIs by definition
        nodeNum = 1
        if roletypes:
            for roleUri in sorted(roletypes.keys()):
                for modelRoleType in roletypes[roleUri]:
                    roleId = modelRoleType.objectId(self.id)
                    node = self.treeView.insert("", "end", modelRoleType.objectId(self.id), text=roleUri, tags=("odd" if nodeNum & 1 else "even",))
                    nodeNum += 1
                    self.treeView.set(node, "definition", modelRoleType.genLabel(lang=self.lang, strip=True) or modelRoleType.definition or '')
                    if self.isArcrole:
                        self.treeView.set(node, "cyclesAllowed", modelRoleType.cyclesAllowed)
                    self.treeView.set(node, "usedOn", ', '.join(str(usedOn)
                                                               for usedOn in modelRoleType.usedOns))

    def getToolTip(self, tvRowId, tvColId):
        # override tool tip when appropriate
        return None

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                # check if modelObject is a relationship in given linkrole
                if isinstance(modelObject, ModelDtsObject.ModelRoleType):
                    roleId = modelObject.objectId()
                else:
                    roleId = None
                # get concept of fact or toConcept of relationship, role obj if roleType
                node = roleId
                if self.treeView.exists(node):
                    self.treeView.see(node)
                    self.treeView.selection_set(node)
            except (AttributeError, KeyError):
                    self.treeView.selection_set(())
            self.blockViewModelObject -= 1



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from enum import Enum
from pathlib import Path

from arelle.ModelValue import QName
from .FilingFormat import FilingFormat


# DEI requirements parsing is designed so that the contents of Attachment #5
# in "(Appendix) Taxonomy Extension Guideline" (ESE140111.zip), or future versions,
# can be easily exported to a CSV file where the rows correspond to DEI concepts
# and the columns correspond to different formats.
class DeiRequirements:
    _csvPath: Path
    _data: dict[QName, dict[FilingFormat, DeiItemStatus | None]] | None
    _deiItems: tuple[QName, ...]
    _filingFormats: tuple[FilingFormat, ...]

    def __init__(self, csvPath: Path, deiItems: tuple[QName, ...], filingFormats: tuple[FilingFormat, ...]):
        self._csvPath = csvPath
        self._data = None
        self._deiItems = deiItems
        self._filingFormats = filingFormats

    def _load(self) -> dict[QName, dict[FilingFormat, DeiItemStatus | None]]:
        if self._data is None:
            with open(self._csvPath, encoding='utf-8') as f:
                data = [
                    [
                        DeiItemStatus.parse(cell) for cell in line.strip().split(',')
                    ]
                    for line in f.readlines()
                ]
                self._data = {}
                assert len(data) == len(self._deiItems), \
                    "Unexpected number of rows in DEI requirements CSV."
                for rowIndex, row in enumerate(data):
                    assert len(row) == len(self._filingFormats), \
                        f"Unexpected number of columns in DEI requirements CSV at row {rowIndex}."
                    deiItem = self._deiItems[rowIndex]
                    self._data[deiItem] = {}
                    for colIndex, cell in enumerate(row):
                        filingFormat = self._filingFormats[colIndex]
                        self._data[deiItem][filingFormat] = cell
        return self._data


    def get(self, deiItem: QName, filingFormat: FilingFormat) -> DeiItemStatus | None:
        data = self._load()
        if deiItem not in data:
            return None
        if filingFormat not in data[deiItem]:
            return None
        return data[deiItem][filingFormat]


class DeiItemStatus(Enum):
    # The values of the enum correspond to the symbols used in the spreadsheet.
    PROHIBITED = '×'
    OPTIONAL = '△'
    CONDITIONAL = '○'
    REQUIRED = '◎'

    @classmethod
    def parse(cls, value: str) -> DeiItemStatus | None:
        try:
            return cls(value)
        except ValueError:
            return None

# The below values are based on Attachment #5 in "(Appendix) Taxonomy Extension Guideline" (ESE140111.zip).
# Column D lists the DEI concepts. Rows purely for grouping are omitted.
# The order is preserved. The index is used to map to other data structures.
DEI_LOCAL_NAMES = (
    # Submitter Information (提出者情報)
    'EDINETCodeDEI',
    'FundCodeDEI',
    'SecurityCodeDEI',
    'FilerNameInJapaneseDEI',
    'FilerNameInEnglishDEI',
    'FundNameInJapaneseDEI',
    'FundNameInEnglishDEI',

    # Document Submission Information (提出書類情報)
    'CabinetOfficeOrdinanceDEI',
    'DocumentTypeDEI',
    'AccountingStandardsDEI',
    'WhetherConsolidatedFinancialStatementsArePreparedDEI',
    'IndustryCodeWhenConsolidatedFinancialStatementsArePreparedInAccordanceWithIndustrySpecificRegulationsDEI',
    'IndustryCodeWhenFinancialStatementsArePreparedInAccordanceWithIndustrySpecificRegulationsDEI',

    # Current Fiscal Year (当会計期間)
    'CurrentFiscalYearStartDateDEI',
    'CurrentPeriodEndDateDEI',
    'TypeOfCurrentPeriodDEI',
    'CurrentFiscalYearEndDateDEI',

    # Previous Fiscal Year (比較対象会計期間)
    'PreviousFiscalYearStartDateDEI',
    'ComparativePeriodEndDateDEI',
    'PreviousFiscalYearEndDateDEI',

    # Next Fiscal Year (次の中間期の会計期間)
    'NextFiscalYearStartDateDEI',
    'EndDateOfQuarterlyOrSemiAnnualPeriodOfNextFiscalYearDEI',

    'NumberOfSubmissionDEI',
    'AmendmentFlagDEI',
    'IdentificationOfDocumentSubjectToAmendmentDEI',

    # Type of Correction (訂正の種類)
    'ReportAmendmentFlagDEI',
    'XBRLAmendmentFlagDEI',
)



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
import os
from arelle import ViewWinTree
from arelle.ModelDtsObject import ModelRelationship
from arelle.ModelInstanceObject import ModelFact
from arelle.ModelVersObject import ModelRelationshipSetChange, ModelInstanceAspectsChange

def viewVersReport(modelXbrl, tabWin):
    modelXbrl.modelManager.showStatus(_("viewing versioning report"))
    view = ViewVersReport(modelXbrl, tabWin)
    view.view()
    # pop up menu
    menu = view.contextMenu()
    menu.add_cascade(label=_("Expand"), underline=0, command=view.expand)
    menu.add_cascade(label=_("Collapse"), underline=0, command=view.collapse)
    # set up treeView widget and tabbed pane
    view.treeView.column("#0", width=300, anchor="w")
    view.treeView.heading("#0", text=_("Versioning Report"))
    view.menuAddClipboard()
    view.menuAddLangs()
    view.menuAddLabelRoles(includeConceptName=True,menulabel=_("Concept Label Role"))

    # sort URIs by definition
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

class ViewVersReport(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin):
        super(ViewVersReport, self).__init__(modelXbrl, tabWin, "Versioning Report", True)

    def view(self):
        self.blockSelectEvent = 1
        self.blockViewModelObject = 0
        self.tag_has = defaultdict(list) # temporary until Tk 8.6

        self.clearTreeView()

        versReport = self.modelXbrl.modelDocument
        # root node for tree view
        self.id = 1
        rootnode = self.treeView.insert("", "end", versReport.objectId(),
                    text=os.path.basename(self.modelXbrl.modelDocument.basename),
                    tags=("odd",))
        nsRenamingsNode = self.treeView.insert("", "end", "node_{0}".format(self.id),
                    text=_("namespace renamings"),
                    tags=("even",))
        self.id += 1
        srtfromURIs = sorted(versReport.namespaceRenameFrom)
        for i, fromURI in enumerate(srtfromURIs):
            nsRenaming = versReport.namespaceRenameFrom[fromURI]
            self.treeView.insert(nsRenamingsNode, "end", nsRenaming.objectId('ns'),
                        text=nsRenaming.viewText(),
                        tags=("even" if i & 1 else "odd",))

        roleChangesNode = self.treeView.insert("", "end", "node_{0}".format(self.id),
                    text=_("role changes"),
                    tags=("odd",))
        self.id += 1
        srtfromURIs = list(versReport.roleChanges.keys())
        srtfromURIs.sort()
        for i, fromRole in enumerate(srtfromURIs):
            roleChange = versReport.roleChanges[fromRole]
            self.treeView.insert(roleChangesNode, "end", roleChange.objectId('role'),
                        text=roleChange.viewText(),
                        tags=("odd" if i & 1 else "even",))

        assignmentsNode = self.treeView.insert("", "end", "node_{}".format(self.id),
                    text=_("assignments"),
                    tags=("even",))
        self.id += 1
        srtAssignmentIds = list(versReport.assignments.keys())
        srtAssignmentIds.sort()
        for i, assignmentId in enumerate(srtAssignmentIds):
            assignment = versReport.assignments[assignmentId]
            text = "{0}: {1} {2}".format(assignmentId,
                                         assignment.genLabel(lang=self.lang) or "",
                                         assignment.categoryQName)
            label = assignment.genLabel(lang=self.lang)
            text = (assignmentId + ": " + label) if label else assignmentId
            self.treeView.insert(assignmentsNode, "end", assignment.objectId(),
                                 text=text,
                                 tags=("even" if i & 1 else "odd",))

        actionsNode = self.treeView.insert("", "end", "node_{0}".format(self.id),
                    text=_("actions"),
                    tags=("odd",))
        self.id += 1
        srtActionIds = list(versReport.actions.keys())
        srtActionIds.sort()
        for i, actionId in enumerate(srtActionIds):
            action = versReport.actions[actionId]
            label = action.genLabel(lang=self.lang)
            text = (actionId + ": " + label) if label else actionId
            actionNode = self.treeView.insert(actionsNode, "end", action.objectId(), text=text,
                                              tags=("odd" if i & 1 else "even",))
            for j, event in enumerate(action.events):
                label = event.genLabel(lang=self.lang)
                text = (event.localName + ": " + label) if label else event.localName
                if isinstance(event, ModelRelationshipSetChange):
                    eventNode = self.treeView.insert(actionNode, "end", event.objectId(), text=text,
                                                     tags=("even" if i+j & 1 else "odd",))
                    k = i + j
                    for relationshipSet, name in ((event.fromRelationshipSet, "fromRelationshipSet"),
                                                  (event.toRelationshipSet, "toRelationshipSet")):
                        if relationshipSet is not None:
                            relSetNode = self.treeView.insert(eventNode, "end", relationshipSet.objectId(),
                                                              text=relationshipSet.localName,
                                                              tags=("odd" if k & 1 else "even",))
                            k += 1
                            l = k
                            for relationship in relationshipSet.relationships:
                                self.treeView.insert(relSetNode, "end", relationship.objectId(), text=relationship.localName,
                                                     tags=("odd" if l & 1 else "even",))
                                l += 1
                elif isinstance(event, ModelInstanceAspectsChange):
                    eventNode = self.treeView.insert(actionNode, "end", event.objectId(), text=text)
                    k = i + j
                    for aspects, name in ((event.fromAspects, "fromAspects"),
                                                  (event.toAspects, "toAspects")):
                        if aspects is not None:
                            k += 1
                            aspectsNode = self.treeView.insert(eventNode, "end", aspects.objectId(), text=aspects.localName,
                                                               tags=("even" if k & 1 else "odd",))
                            l = k
                            for aspect in aspects.aspects:
                                l += 1
                                aspectNode = self.treeView.insert(aspectsNode, "end", aspect.objectId(),
                                                                  text=aspect.localName + " " + aspect.elementAttributesStr,
                                                                  tags=("even" if l & 1 else "odd",))
                                m = l
                                for member in getattr(aspect, "relatedConcepts", getattr(aspect, "relatedPeriods", getattr(aspect, "relatedMeasures", []))):
                                    self.treeView.insert(aspectNode, "end", member.objectId(),
                                                         text=member.localName + " " + member.elementAttributesStr,
                                                         tags=("even" if m & 1 else "odd",))
                else:
                    self.treeView.insert(actionNode, "end", event.objectId(),
                                         text=text + " " + event.viewText(self.labelrole, self.lang),
                                         tags=("even" if i+j & 1 else "odd",))

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                if isinstance(modelObject, ModelRelationship):
                    conceptId = modelObject.toModelObject.objectId()
                elif isinstance(modelObject, ModelFact):
                    conceptId = self.modelXbrl.qnameConcepts[modelObject.qname].objectId()
                else:
                    conceptId = modelObject.objectId()
                    '''
                items = self.treeView.tag_has(conceptId)
                    '''
                items = self.tag_has.get(conceptId)
                if items is not None and self.treeView.exists(items[0]):
                    self.treeView.see(items[0])
                    self.treeView.selection_set(items[0])
            except KeyError:
                    self.treeView.selection_set(())
            self.blockViewModelObject -= 1



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Any, Optional, Union

from arelle.ModelValue import QName
from arelle.PythonUtil import OrderedSet
from .XbrlTaxonomyObject import XbrlReferencableTaxonomyObject
from .XbrlTypes import XbrlTaxonomyType, QNameKeyType, DefaultFalse

class XbrlProperty:
    property: QName # (required) The name is a QName that uniquely identifies the property type object.
    value: Any # (required) The value of the property, that must be consistent with the datatype of the property.

    @property
    def propertyView(self):
        return ( str(getattr(self, "property", "")), str(getattr(self, "value", "")) )


class XbrlPropertyType(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The name is a QName that uniquely identifies the property type object.
    dataType: QName # (required) Indicates the dataType of the property value. These are provided as a QName based on the datatypes specified in the XBRL 2.1 specification and any custom datatype defined in the taxonomy.
    enumerationDomain: Optional[QName] # (optional) Used to specify the QName of a domain object that is used to derive enumerated domain members QNames that can be used for the property.
    definitional: Union[bool, DefaultFalse] # (optional) Indicates if the property is definitional. If changes to the property change the meaning of the object it is definitional, if the property provides extra information about the object it is not definitional. If no value is provided the attribute defaults to false.
    allowedObjects: set[QName] # (optional) List of allowable objects that the property can be used with. For example the balance property can only be used with concept objects.
    allowedAsLinkProperty: Union[bool, DefaultFalse] # (optional) Indicates if the property can be used a a properton the link between two objects in a relationship. If no value is provided the attribute defaults to false.



'''
See COPYRIGHT.md for copyright information.
'''
from arelle.ModelDtsObject import ModelConcept
from arelle.ModelInstanceObject import ModelFact
from arelle.ModelObject import ModelObject
from arelle.PythonUtil import flattenSequence
from arelle.Version import authorLabel, copyrightLabel
from arelle.XmlUtil import xmlstring, descendantAttr
from arelle import XbrlConst, XmlUtil
import regex as re
from collections import defaultdict
from decimal import Decimal

labelrole = None
lang = None

altParametersPattern = re.compile(r"\$\{([\w.:-]+)\}")
parametersPattern = re.compile(r"%\(([\w.:-]+)\)")
factNumberPattern = re.compile(r"fact(\d+)")

def measureFormat(measure):
    if measure.namespaceURI in (XbrlConst.iso4217, XbrlConst.xbrli):
        return measure.localName
    return str(measure)  # qname str

def loggingMessageParameters(messageCode, msgIn, modelObjectArgs, fmtArgs, *args, **kwargs):
    if messageCode and messageCode.startswith("DQC"):
        # change ${...} in message into %(...)s
        msg = altParametersPattern.sub(r"%(\1)s", msgIn)

        # find qnamed fact references
        qnamedReferences = set()
        paramNames = parametersPattern.findall(msg)
        for param in paramNames:
            paramParts = param.split('.')
            if len(paramParts) > 1 and not factNumberPattern.match(paramParts[0]):
                qnamedReferences.add(paramParts[0])
        factsArray = []
        factsByQname = defaultdict(list)  # list of facts with the qname
        conceptsByQname = {}
        missingQnamedArguments = set()
        qnameFirstOrdinalPosition = {}  # for qnamed facts multiply in sequence (to insert on 2nd occurance in ordinal position)
        for arg in flattenSequence(modelObjectArgs):
            # Pargmeter may be a ModelFact, or a name of a concept (such as a dimension)
            if isinstance(arg, ModelFact):
                _strQName = str(arg.qname)
                if _strQName in qnamedReferences:
                    if _strQName not in factsByQname: # if twice in args, let one be an ordinal fact
                        factsByQname[_strQName].append(arg)
                        qnameFirstOrdinalPosition[_strQName] = len(factsArray)
                    elif arg not in factsArray: # if twice in args, insert first occurence in ordinal position
                        factsArray.insert(qnameFirstOrdinalPosition[_strQName], arg)
                else:
                    factsArray.append(arg)
                cntx = arg.context
                if cntx is not None:
                    for dim in cntx.qnameDims.values():
                        if str(dim.dimensionQname) in qnamedReferences:
                            conceptsByQname[str(dim.dimensionQname)] = dim.dimension
                        elif str(dim.dimensionQname) in qnamedReferences:
                            conceptsByQname[str(dim.memberQname)] = dim.member

        def setArgForFactProperty(param, modelFact, propertyNameParts):
            propVal = None
            property = propertyNameParts[0]
            if property == "value":
                if isinstance(modelFact.xValue, Decimal):
                    propVal = "{:,}".format(modelFact.xValue)
                else:
                    propVal = modelFact.value
            elif property == "decimals":
                propVal = modelFact.decimals
            elif property == "label" and modelFact.concept is not None:
                propVal = modelFact.concept.label(labelrole,
                                                  lang=lang,
                                                  linkroleHint=XbrlConst.defaultLinkRole)
            elif property == "name":
                propVal = str(modelFact.qname)
            elif property == "localName":
                propVal = modelFact.qname.localName
            else:
                cntx = modelFact.context
                unit = modelFact.unit
                if cntx is not None:
                    if property == "period":
                        if len(propertyNameParts) == 1:
                            if cntx.isForeverPeriod:
                                propVal = "forever"
                            elif cntx.isInstantPeriod:
                                propVal = XmlUtil.dateunionValue(cntx.instantDatetime, subtractOneDay=True)
                            else:
                                propVal = "{} to {}".format(XmlUtil.dateunionValue(cntx.startDatetime),
                                                           XmlUtil.dateunionValue(cntx.endDatetime, subtractOneDay=True))
                        else:
                            dateSelection = propertyNameParts[1]
                            if dateSelection == "startDate":
                                propVal = XmlUtil.dateunionValue(cntx.startDatetime)
                            elif dateSelection == "endDate":
                                propVal = XmlUtil.dateunionValue(cntx.endDatetime, subtractOneDay=True)
                            elif dateSelection == "instant":
                                propVal = XmlUtil.dateunionValue(cntx.instant, subtractOneDay=True)
                            elif dateSelection == "durationDays":
                                propVal = str((cntx.endDatetime - cntx.startDatetime).days)
                    elif property == "dimensions":
                        if cntx.qnameDims:
                            propVal = "\n".join("{} = {}".format(d.dimensionQname,
                                                                d.memberQname if d.isExplicit else
                                                                XmlUtil.xmlstring( XmlUtil.child(d), stripXmlns=True, prettyPrint=True ))
                                                for d in cntx.qnameDims.values())
                        else:
                            propVal = "none"
                if property == "unit":
                    if unit is None:
                        propVal = "none"
                    else:
                        measures = unit.measures
                        if measures[1]:
                            propVal = 'mul {}\ndiv {} '.format(
                                    ', '.join(measureFormat(m) for m in measures[0]),
                                    ', '.join(measureFormat(m) for m in measures[1]))
                        else:
                            propVal = ', '.join(measureFormat(m) for m in measures[0])
            fmtArgs[param] = propVal

        def setArgForConceptProperty(param, modelConceptOrQname, propertyNameParts):
            propVal = None
            property = propertyNameParts[0]
            if property == "label":
                if isinstance(modelConceptOrQname, ModelConcept):
                    propVal = modelConceptOrQname.label(labelrole,
                                                        lang=lang,
                                                        linkroleHint=XbrlConst.defaultLinkRole)
                elif isinstance(modelConceptOrQname, str):
                    propVal = modelConceptOrQname
            elif property == "name":
                if isinstance(modelConceptOrQname, ModelConcept):
                    propVal = str(modelConceptOrQname.qname)
                elif isinstance(modelConceptOrQname, str):
                    propVal = modelConceptOrQname
            elif property == "localName":
                if isinstance(modelConcept, ModelConcept):
                    propVal = modelConcept.qname.localName
                elif isinstance(modelConcept, str):
                    propVal = modelConcept.rpartition(':')[2]
            if propVal is not None:
                fmtArgs[param] = propVal

        # parse parameter names out of msg
        for param in paramNames:
            try:
                paramParts = param.split('.')
                if len(paramParts) >= 2 and factNumberPattern.match(paramParts[0]):
                    modelFactNum = int(factNumberPattern.match(paramParts[0]).group(1))
                    if 1 <= modelFactNum <= len(factsArray):
                        modelFact = factsArray[modelFactNum - 1]
                        setArgForFactProperty(param, modelFact, paramParts[1:])
                elif len(paramParts) >= 3 and paramParts[1] == "fact":
                    # take first matching fact ??
                    if paramParts[0] in factsByQname:
                        modelFact = factsByQname[paramParts[0]][0] # there may be multiple facts of this QName, take first
                        setArgForFactProperty(param, modelFact, paramParts[2:])
                    else:
                        missingQnamedArguments.add(paramParts[0])
                elif len(paramParts) >= 2:
                    modelConcept = conceptsByQname.get(paramParts[0], paramParts[0]) # if no model concept pass in the qname
                    setArgForConceptProperty(param, modelConcept, paramParts[1:])
            except Exception as _ex:
                raise # pass
        if missingQnamedArguments:
            for arg in modelObjectArgs:
                # Pargmeter may be a ModelFact, or a name of a concept (such as a dimension)
                if isinstance(arg, ModelObject):
                    arg.modelXbrl.error("dqcErrorLog:unresolvedParameters",
                                        _("The error message %(messageCode)s has unresolved named parameters: %(unresolvedParameters)s"),
                                        modelObject=modelObjectArgs, messageCode=messageCode,
                                        unresolvedParameters=', '.join(sorted(missingQnamedArguments)))
                    break
            for missingArgument in missingQnamedArguments:
                fmtArgs[missingArgument] = "unavailable"
        return msg
    return None

def loggingCommandLineXbrlRun(cntlr, options, modelXbrl, *args, **kwargs):
    global labelrole, lang
    labelrole=options.labelRole
    lang=options.labelLang

def testcaseVariationExpectedSeverity(modelTestcaseVariation, *args, **kwargs):
    _severity = descendantAttr(modelTestcaseVariation, None, "error", "severity")
    if _severity is not None:
        return _severity.upper()
    return None

def testcaseVariationExpectedCount(modelTestcaseVariation, *args, **kwargs):
    try:
        return int(descendantAttr(modelTestcaseVariation, None, "error", "count"))
    except (ValueError, TypeError):
        pass
    return None

__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Logging - DQC Parameters',
    'version': '1.0',
    'description': '''DQC tests logging messages: adds parameter values from infrastructurally
provided logging arguments.  Usually uses modelObject arguments to supply parameters found
in message text that can be derived from the arguments.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrCmdLine.Xbrl.Run': loggingCommandLineXbrlRun,
    'Logging.Message.Parameters': loggingMessageParameters,
    'ModelTestcaseVariation.ExpectedSeverity': testcaseVariationExpectedSeverity,
    'ModelTestcaseVariation.ExpectedCount': testcaseVariationExpectedCount
}
